# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import math
import numpy as np
import quaternion
import spinsfast
from numpy import *
import pytest

from conftest import linear_waveform, constant_waveform, random_waveform


@pytest.mark.parametrize("w", [linear_waveform, constant_waveform])
def test_dpa_simple_cases(w):
    from scri.mode_calculations import LLDominantEigenvector

    LL = LLDominantEigenvector(w())
    LL_expected = np.zeros_like(LL, dtype=float)
    LL_expected[:, 2] = 1.0
    assert np.allclose(LL, LL_expected)


@pytest.mark.parametrize("w", [linear_waveform, constant_waveform])
def test_dpa_rotated_simple_cases(w, Rs):
    from scri.mode_calculations import LLDominantEigenvector

    # We use `begin=1.0` because we need to avoid situations where the modes
    # are all zeros, which can happen in `linear_waveform` at t=0.0
    W = w(begin=1.0, ell_min=0, n_times=len(Rs))
    W.rotate_physical_system(Rs)
    LL = LLDominantEigenvector(W)
    LL_expected = quaternion.as_float_array(Rs * np.array([quaternion.z for i in range(len(Rs))]) * (~Rs))[:, 1:]

    # Because the dpa is only defined up to a sign, all we need is for the
    # dot product between the dpa and the expected value to be close to
    # either 1 or -1.  This finds the largest difference, based on the
    # smaller of the two sign options.
    assert (
        max(
            np.amin(
                np.vstack((np.linalg.norm(LL - LL_expected, axis=1), np.linalg.norm(LL + LL_expected, axis=1))), axis=0
            )
        )
        < 1.0e-14
    )


@pytest.mark.parametrize("w", [linear_waveform, constant_waveform, random_waveform])
def test_dpa_rotated_generally(w, Rs):
    from scri.mode_calculations import LLDominantEigenvector

    n_copies = 10
    W = w(begin=1.0, end=100.0, n_times=n_copies * len(Rs), ell_min=0, ell_max=8)
    assert W.ensure_validity(alter=False)
    R_basis = np.array([R for R in Rs for i in range(n_copies)])

    # We use `begin=1.0` because we need to avoid situations where the modes
    # are all zeros, which can happen in `linear_waveform` at t=0.0
    LL1 = LLDominantEigenvector(W)
    LL1 = quaternion.as_float_array(np.array([R * np.quaternion(0, *v) * (~R) for R, v in zip(R_basis, LL1)]))[:, 1:]
    W.rotate_physical_system(R_basis)
    LL2 = LLDominantEigenvector(W)

    # Because the dpa is only defined up to a sign, all we need is for the
    # dot product between the dpa and the expected value to be close to
    # either 1 or -1.  This finds the largest difference, based on the
    # smaller of the two sign options.
    assert (
        max(np.amin(np.vstack((np.linalg.norm(LL1 - LL2, axis=1), np.linalg.norm(LL1 + LL2, axis=1))), axis=0))
        < 1.0e-12
    )


def test_zero_angular_velocity():
    from scri.mode_calculations import angular_velocity

    w = constant_waveform(end=10.0, n_times=10000)
    Omega_out = angular_velocity(w)
    assert np.allclose(Omega_out, np.zeros_like(Omega_out), atol=1e-15, rtol=0.0)


def test_z_angular_velocity():
    from scri.mode_calculations import angular_velocity

    w = constant_waveform(end=10.0, n_times=10000)
    omega = 2 * math.pi / 5.0
    R = np.exp(quaternion.quaternion(0, 0, 0, omega / 2) * w.t)
    w.rotate_physical_system(R)
    Omega_out = angular_velocity(w)
    Omega_in = np.zeros_like(Omega_out)
    Omega_in[:, 2] = omega
    assert np.allclose(Omega_in, Omega_out, atol=1e-12, rtol=2e-8)


def test_rotated_angular_velocity():
    from scri.mode_calculations import angular_velocity

    w = constant_waveform(end=10.0, n_times=10000)
    omega = 2 * math.pi / 5.0
    R0 = quaternion.quaternion(1, 2, 3, 4).normalized()
    R = R0 * np.exp(quaternion.quaternion(0, 0, 0, omega / 2) * w.t)
    w.rotate_physical_system(R)
    Omega = R0 * quaternion.quaternion(0, 0, 0, omega) * R0.inverse()
    Omega_out = angular_velocity(w)
    Omega_in = np.zeros_like(Omega_out)
    Omega_in[:, 0] = Omega.x
    Omega_in[:, 1] = Omega.y
    Omega_in[:, 2] = Omega.z
    assert np.allclose(Omega_in, Omega_out, atol=1e-12, rtol=2e-8)


def test_corotating_frame():
    from scri.mode_calculations import corotating_frame
    from scri import Corotating

    w = constant_waveform(end=10.0, n_times=100000)  # Need lots of time steps for accurate integration
    omega = 2 * math.pi / 5.0
    R0 = quaternion.quaternion(1, 2, 3, 4).normalized()
    R_in = R0 * np.exp(quaternion.quaternion(0, 0, 0, omega / 2) * w.t)
    w_rot = w.deepcopy()
    w_rot.rotate_physical_system(R_in)
    R_out = corotating_frame(w_rot, R0=R0, tolerance=1e-12)
    assert np.allclose(quaternion.as_float_array(R_in), quaternion.as_float_array(R_out), atol=1e-10, rtol=0.0)
    w_rot.to_corotating_frame(R0=R0, tolerance=1e-12)
    assert w._allclose(w_rot, atol=1e-8)
    assert w_rot.frameType == Corotating



import math
import numpy as np
import quaternion
import spinsfast
import spherical_functions as sf
import scri
import pytest

ABD = scri.AsymptoticBondiData
abd = scri.asymptotic_bondi_data


def construct_and_validate(modifier, validator, ell_max=8):
    time = np.linspace(-100, 100, num=2001)
    sigma, sigmadot, sigmaddot, psi2, psi1, psi0 = np.zeros((6, sf.LM_total_size(0, ell_max)), dtype=complex)
    modifier(sigma, sigmadot, sigmaddot, psi2, psi1, psi0)
    abd = ABD.from_initial_values(time, ell_max, sigma, sigmadot, sigmaddot, psi2, psi1, psi0)
    validator(abd)
    return True


def check_modes(modes, nonzero_ℓm):
    import numpy as np
    import spherical_functions as sf

    non_zero_indices = np.array([modes.index(ℓ, m) for ℓ, m in nonzero_ℓm], dtype=int)
    zero_indices = np.array(list(set(range(sf.LM_total_size(0, modes.ell_max))) - set(non_zero_indices)), dtype=int)
    assert not np.any(modes[..., zero_indices]), f"nonzero values among indices {zero_indices}"
    for non_zero_index in non_zero_indices:
        assert np.any(modes[..., non_zero_index]), f"no nonzero values at index {non_zero_index}"


def test0():
    """Set only terms that are forbidden by spin weights; ensure all zeros"""

    def modifier(sigma, sigmadot, sigmaddot, psi2, psi1, psi0):
        psi0[: sf.LM_index(1, 1, 0)] = 1.234
        psi1[0] = 0.123
        sigma[: sf.LM_index(1, 1, 0)] = 0.567
        sigmadot[: sf.LM_index(1, 1, 0)] = 0.678
        sigmaddot[: sf.LM_index(1, 1, 0)] = 0.789

    def validator(abd):
        check_modes(abd.psi0, [])
        check_modes(abd.psi1, [])
        check_modes(abd.psi2, [])
        check_modes(abd.psi3, [])
        check_modes(abd.psi4, [])
        check_modes(abd.sigma, [])
        assert np.max(np.abs(abd.bondi_violation_norms)) == 0.0

    construct_and_validate(modifier, validator)


def test1():
    """Add ℓ=0 term to ψ₂ intial value

    Ensures that first terms of ψ̇₁u = ðψ₂ + 2σψ₃ and ψ̇₀ = ðψ₁ + 3σψ₂ don't get excited

    """

    def modifier(sigma, sigmadot, sigmaddot, psi2, psi1, psi0):
        # Nonsensical values that should have no effect
        psi0[: sf.LM_index(1, 1, 0)] = 1.234
        psi1[0] = 0.123
        sigma[: sf.LM_index(1, 1, 0)] = 0.567
        sigmadot[: sf.LM_index(1, 1, 0)] = 0.678
        sigmaddot[: sf.LM_index(1, 1, 0)] = 0.789
        # Actual values that should carry through
        psi2[sf.LM_index(0, 0, 0)] = 0.234

    def validator(abd):
        assert np.all(abd.psi2[..., 0] == 0.234)
        check_modes(abd.psi0, [])
        check_modes(abd.psi1, [])
        check_modes(abd.psi2, [[0, 0]])
        check_modes(abd.psi3, [])
        check_modes(abd.psi4, [])
        check_modes(abd.sigma, [])
        assert np.max(np.abs(abd.bondi_violation_norms)) == 0.0

    construct_and_validate(modifier, validator, ell_max=3)


def test2():
    """Add ℓ=1 term to ψ₂ intial value

    Checks first term of ψ̇₁ = ðψ₂ + 2σψ₃

    """

    def modifier(sigma, sigmadot, sigmaddot, psi2, psi1, psi0):
        # Nonsensical values that should have no effect
        psi0[: sf.LM_index(1, 1, 0)] = 1.234
        psi1[0] = 0.123
        sigma[: sf.LM_index(1, 1, 0)] = 0.567
        sigmadot[: sf.LM_index(1, 1, 0)] = 0.678
        sigmaddot[: sf.LM_index(1, 1, 0)] = 0.789
        # Actual values that should carry through
        psi2[sf.LM_index(0, 0, 0)] = 0.234
        psi2[sf.LM_index(1, -1, 0)] = 0.345

    def validator(abd):
        check_modes(abd.psi0, [])
        check_modes(abd.psi1, [[1, -1], [1, 1]])
        check_modes(abd.psi2, [[0, 0], [1, -1], [1, 1]])
        check_modes(abd.psi3, [])
        check_modes(abd.psi4, [])
        check_modes(abd.sigma, [])
        assert np.max(np.abs(abd.bondi_violation_norms)) < 1e-13

    construct_and_validate(modifier, validator, ell_max=4)


def test3():
    """Add ℓ=2 term to ψ₂ intial value

    Checks first term of ψ̇₀ = ðψ₁ + 3σψ₂

    """

    def modifier(sigma, sigmadot, sigmaddot, psi2, psi1, psi0):
        # Nonsensical values that should have no effect
        psi0[: sf.LM_index(1, 1, 0)] = 1.234
        psi1[0] = 0.123
        sigma[: sf.LM_index(1, 1, 0)] = 0.567
        sigmadot[: sf.LM_index(1, 1, 0)] = 0.678
        sigmaddot[: sf.LM_index(1, 1, 0)] = 0.789
        # Actual values that should carry through
        psi2[sf.LM_index(0, 0, 0)] = 0.234
        psi2[sf.LM_index(1, -1, 0)] = 0.345
        psi2[sf.LM_index(2, -2, 0)] = 0.456

    def validator(abd):
        assert np.all(abd.psi2[..., 0] == 0.234)
        check_modes(abd.psi0, [[2, -2], [2, 2]])
        check_modes(abd.psi1, [[1, -1], [1, 1], [2, -2], [2, 2]])
        check_modes(abd.psi2, [[0, 0], [1, -1], [1, 1], [2, -2], [2, 2]])
        check_modes(abd.psi3, [])
        check_modes(abd.psi4, [])
        check_modes(abd.sigma, [])
        assert np.max(np.abs(abd.bondi_violation_norms)) < 4e-11

    construct_and_validate(modifier, validator, ell_max=4)


def test4():
    """Add nonzero constant term to σ

    Checks first term of ψ̇₁ = ðψ₂ + 2σψ₃ and second term of ψ̇₀ = ðψ₁ + 3σψ₂

    After satisfaction of the reality condition on the mass aspect, ψ₂ has nonzero modes in
    {(0, 0), (2, -2), (2, 2)}, so ψ₁ should have nonzero modes in {(2, -2), (2, 2)}.  Here,
    σ has only the nonzero mode (2, 2).  Thus the product σψ₂ should result in nonzero
    modes in {(2, -2), (2, 0), (2, 2), (3, 0), (4, 0), (4, 4)}.

    """

    def modifier(sigma, sigmadot, sigmaddot, psi2, psi1, psi0):
        # Nonsensical values that should have no effect
        psi0[: sf.LM_index(1, 1, 0)] = 1.234
        psi1[0] = 0.123
        sigma[: sf.LM_index(1, 1, 0)] = 0.567
        sigmadot[: sf.LM_index(1, 1, 0)] = 0.678
        sigmaddot[: sf.LM_index(1, 1, 0)] = 0.789
        # Actual values that should carry through
        psi2[0] = 0.234
        sigma[sf.LM_index(2, 2, 0)] = 0.5678

    def validator(abd):
        check_modes(abd.psi0, [[2, -2], [2, 0], [2, 2], [3, 0], [4, 0], [4, 4]])
        check_modes(abd.psi1, [[2, -2], [2, 2]])
        check_modes(abd.psi2, [[0, 0], [2, -2], [2, 2]])
        check_modes(abd.psi3, [])
        check_modes(abd.psi4, [])
        check_modes(abd.sigma, [[2, 2]])
        assert np.max(np.abs(abd.bondi_violation_norms)) <= 2e-10

    construct_and_validate(modifier, validator, ell_max=6)


def test5():
    """Add nonzero ℓ=2 term to σ̇

    Checks second term of ψ̇₁ = ðψ₂ + 2σψ₃

    """

    def modifier(sigma, sigmadot, sigmaddot, psi2, psi1, psi0):
        # Nonsensical values that should have no effect
        psi0[: sf.LM_index(1, 1, 0)] = 1.234
        psi1[0] = 0.123
        sigma[: sf.LM_index(1, 1, 0)] = 0.567
        sigmadot[: sf.LM_index(1, 1, 0)] = 0.678
        sigmaddot[: sf.LM_index(1, 1, 0)] = 0.789
        # Actual values that should carry through
        sigma[sf.LM_index(2, 2, 0)] = 0.5678
        sigmadot[sf.LM_index(2, 2, 0)] = 0.6789

    def validator(abd):
        check_modes(abd.psi0, [[2, -2], [2, 0], [2, 2], [3, 0], [4, 0], [4, 4]])
        check_modes(abd.psi1, [[1, 0], [2, -2], [2, 0], [2, 2], [3, 0], [4, 0]])
        check_modes(abd.psi2, [[2, -2], [2, 2]])
        check_modes(abd.psi3, [[2, -2]])
        check_modes(abd.psi4, [])
        check_modes(abd.sigma, [[2, 2]])
        assert np.max(np.abs(abd.bondi_violation_norms)) <= 7e-9

    construct_and_validate(modifier, validator, ell_max=6)


def test6():
    """Add nonzero ℓ=2 term to σ̈

    Checks second term of ψ̇₂ = ðψ₃ + σψ₄

    """

    def modifier(sigma, sigmadot, sigmaddot, psi2, psi1, psi0):
        # Nonsensical values that should have no effect
        psi0[: sf.LM_index(1, 1, 0)] = 1.234
        psi1[0] = 0.123
        sigma[: sf.LM_index(1, 1, 0)] = 0.567
        sigmadot[: sf.LM_index(1, 1, 0)] = 0.678
        sigmaddot[: sf.LM_index(1, 1, 0)] = 0.789
        # Actual values that should carry through
        sigmaddot[sf.LM_index(2, 2, 0)] = 0.1 / 10_000 ** 2  # 10_000 = max(time)**2

    def validator(abd):
        check_modes(abd.psi0, [[2, -2], [2, 0], [2, 2], [3, 0], [3, 2], [4, 0], [4, 2], [5, 2], [6, 2]])
        check_modes(abd.psi1, [[1, 0], [2, -2], [2, 0], [3, 0], [4, 0]])
        check_modes(abd.psi2, [[0, 0], [1, 0], [2, -2], [2, 0], [3, 0], [4, 0]])
        check_modes(abd.psi3, [[2, -2]])
        check_modes(abd.psi4, [[2, -2]])
        check_modes(abd.sigma, [[2, 2]])
        assert np.max(np.abs(abd.bondi_violation_norms)) <= 5e-8

    construct_and_validate(modifier, validator, ell_max=7)


def test7():
    """Test random values for all ℓ modes"""
    ell_max = 8
    np.random.seed(1234)

    def modifier(sigma, sigmadot, sigmaddot, psi2, psi1, psi0):
        sigma[:] = 0.01 * np.random.rand(*(sigma.shape[:-1] + (sigma.shape[-1] * 2,))).view(complex)
        sigmadot[:] = (0.01 / 100) * np.random.rand(*(sigma.shape[:-1] + (sigma.shape[-1] * 2,))).view(complex)
        sigmaddot[:] = (0.01 / 100 ** 2) * np.random.rand(*(sigma.shape[:-1] + (sigma.shape[-1] * 2,))).view(complex)

    def validator(abd):
        check_modes(abd.psi0, sf.LM_range(abs(abd.psi0.s), ell_max))
        check_modes(abd.psi1, sf.LM_range(abs(abd.psi1.s), ell_max))
        check_modes(abd.psi2, sf.LM_range(abs(abd.psi2.s), ell_max))
        check_modes(abd.psi3, sf.LM_range(2, ell_max))
        check_modes(abd.psi4, sf.LM_range(abs(abd.psi4.s), ell_max))
        check_modes(abd.sigma, sf.LM_range(abs(abd.sigma.s), ell_max))
        assert np.max(np.abs(abd.bondi_violation_norms)) <= 4.5e-6

    construct_and_validate(modifier, validator, ell_max=ell_max)



# Copyright (c) 2019, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import numpy as np
import quaternion
import spinsfast
import scri
from numpy import *
import pytest

from conftest import linear_waveform, constant_waveform, random_waveform


def silly_momentum_flux(h):
    """Compute momentum flux from waveform with a silly but simple method

    This function evaluates the momentum-flux formula quite literally.  The formula is

         dp       R**2  | dh |**2 ^
    ---------- = ------ | -- |    n
    dOmega  dt   16  pi | dt |

    Here, p and nhat are vectors and R is the distance to the source.  The input h is differentiated
    numerically to find modes of dh/dt, which are then used to construct the values of dh/dt on a
    grid.  At each point of that grid, the value of |dh/dt|**2 nhat is computed, which is then
    integrated over the sphere to arrive at dp/dt.  Note that this integration is accomplished by
    way of a spherical-harmonic decomposition; the ell=m=0 mode is multiplied by 2*sqrt(pi) to
    arrive at the result that would be achieved by integrating over the sphere.

    """
    hdot = h.data_dot
    zeros = np.zeros((hdot.shape[0], 4), dtype=hdot.dtype)
    data = np.concatenate((zeros, hdot), axis=1)  # Pad with zeros for spinsfast
    ell_min = 0
    ell_max = 2 * h.ell_max + 1  # Maximum ell value required for nhat*|hdot|^2
    n_theta = 2 * ell_max + 1
    n_phi = n_theta
    hdot_map = spinsfast.salm2map(data, h.spin_weight, h.ell_max, n_theta, n_phi)
    hdot_mag_squared_map = hdot_map * hdot_map.conjugate()

    theta = np.linspace(0.0, np.pi, num=n_theta, endpoint=True)
    phi = np.linspace(0.0, 2 * np.pi, num=n_phi, endpoint=False)
    x = np.outer(np.sin(theta), np.cos(phi))
    y = np.outer(np.sin(theta), np.sin(phi))
    z = np.outer(np.cos(theta), np.ones_like(phi))

    pdot = np.array(
        [
            spinsfast.map2salm(hdot_mag_squared_map * n / (16 * np.pi), 0, 0)[..., 0].real * (2 * np.sqrt(np.pi))
            for n in [x, y, z]
        ]
    ).T

    return pdot


def silly_angular_momentum_flux(h, hdot=None):
    """Compute angular momentum flux from waveform explicitly

    This function uses very explicit (and slow) methods, but different as far as possible from the
    methods used in the main code.

    """
    hdot = hdot or h.data_dot
    zeros = np.zeros((hdot.shape[0], 4), dtype=hdot.dtype)
    data = np.concatenate((zeros, hdot), axis=1)  # Pad with zeros for spinsfast
    ell_min = 0
    ell_max = 2 * h.ell_max  # Maximum ell value required to fully resolve Ji{h} * hdot
    n_theta = 2 * ell_max + 1
    n_phi = n_theta
    hdot_map = spinsfast.salm2map(data, h.spin_weight, h.ell_max, n_theta, n_phi)

    jdot = np.zeros((h.n_times, 3), dtype=float)

    # Compute J_+ h
    for ell in range(h.ell_min, h.ell_max + 1):
        i = h.index(ell, -ell) + 4
        data[:, i] = 0.0j
        for m in range(-ell, ell):
            i = h.index(ell, m + 1) + 4
            j = h.index(ell, m)
            data[:, i] = 1.0j * np.sqrt((ell - m) * (ell + m + 1)) * h.data[:, j]
    jplus_h_map = spinsfast.salm2map(data, h.spin_weight, h.ell_max, n_theta, n_phi)

    # Compute J_- h
    for ell in range(h.ell_min, h.ell_max + 1):
        for m in range(-ell + 1, ell + 1):
            i = h.index(ell, m - 1) + 4
            j = h.index(ell, m)
            data[:, i] = 1.0j * np.sqrt((ell + m) * (ell - m + 1)) * h.data[:, j]
        i = h.index(ell, ell) + 4
        data[:, i] = 0.0j
    jminus_h_map = spinsfast.salm2map(data, h.spin_weight, h.ell_max, n_theta, n_phi)

    # Combine jplus and jminus to compute x-y components, then conjugate, multiply by hdot, and integrate
    jx_h_map = 0.5 * (jplus_h_map + jminus_h_map)
    jy_h_map = -0.5j * (jplus_h_map - jminus_h_map)
    jdot[:, 0] = (
        -spinsfast.map2salm(jx_h_map.conjugate() * hdot_map, 0, 0)[..., 0].real * (2 * np.sqrt(np.pi)) / (16 * np.pi)
    )
    jdot[:, 1] = (
        -spinsfast.map2salm(jy_h_map.conjugate() * hdot_map, 0, 0)[..., 0].real * (2 * np.sqrt(np.pi)) / (16 * np.pi)
    )

    # Compute J_z h, then conjugate, multiply by hdot, and integrate
    for ell in range(h.ell_min, h.ell_max + 1):
        for m in range(-ell, ell + 1):
            i = h.index(ell, m)
            data[:, i + 4] = 1.0j * m * h.data[:, i]
    jz_h_map = spinsfast.salm2map(data, h.spin_weight, h.ell_max, n_theta, n_phi)
    jdot[:, 2] = (
        -spinsfast.map2salm(jz_h_map.conjugate() * hdot_map, 0, 0)[..., 0].real * (2 * np.sqrt(np.pi)) / (16 * np.pi)
    )

    return jdot


def test_momentum_flux():
    h = scri.sample_waveforms.fake_precessing_waveform(t_1=1_000.0)
    pdot1 = silly_momentum_flux(h)
    pdot2 = scri.momentum_flux(h)
    assert np.allclose(pdot1, pdot2, rtol=1e-13, atol=1e-13)


def test_angular_momentum_flux():
    h = scri.sample_waveforms.fake_precessing_waveform(t_1=1_000.0)
    jdot1 = silly_angular_momentum_flux(h)
    jdot2 = scri.angular_momentum_flux(h)
    assert np.allclose(jdot1, jdot2, rtol=1e-13, atol=1e-13)
    

def test_boost_flux():
    from quaternion import rotate_vectors    

    h = scri.sample_waveforms.single_mode(ell=8, m=5)  # Any mode can be used to implement this
    R = np.quaternion(1, 4, 3, 2).normalized()
    A = h.boost_flux()
    for i in range(len(A)):
        A[i, :] = rotate_vectors(R, A[i,:], axis=-1)
    B = (h.rotate_decomposition_basis(~R)).boost_flux()
    assert np.allclose(A, B, rtol=1e-13, atol=1e-13)



# Copyright (c) 2019, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import functools
import numpy as np
from . import jit

maxexp = np.finfo(float).maxexp * np.log(2) * 0.99


@jit
def _transition_function(x, x0, x1, y0, y1):
    transition = np.empty_like(x)
    ydiff = y1 - y0
    i = 0
    while x[i] <= x0:
        i += 1
    i0 = i
    transition[:i0] = y0
    while x[i] < x1:
        tau = (x[i] - x0) / (x1 - x0)
        exponent = 1.0 / tau - 1.0 / (1.0 - tau)
        if exponent >= maxexp:
            transition[i] = y0
        else:
            transition[i] = y0 + ydiff / (1.0 + np.exp(exponent))
        i += 1
    i1 = i
    transition[i1:] = y1
    return transition, i0, i1


def transition_function(x, x0, x1, y0=0.0, y1=1.0, return_indices=False):
    """Return a smooth function that is constant outside (x0, x1).

    This uses the standard smooth (C^infinity) function with derivatives of compact support to
    transition between the two values, being constant outside of the transition region (x0, x1).

    Parameters
    ==========
    x: array_like
        One-dimensional monotonic array of floats.
    x0: float
        Value before which the output will equal `y0`.
    x1: float
        Value after which the output will equal `y1`.
    y0: float [defaults to 0.0]
        Value of the output before `x0`.
    y1: float [defaults to 1.0]
        Value of the output after `x1`.
    return_indices: bool [defaults to False]
        If True, return the array and the indices (i0, i1) at which the transition occurs, such that
        t[:i0]==y0 and t[i1:]==y1.

    """
    if return_indices:
        return _transition_function(x, x0, x1, y0, y1)
    return _transition_function(x, x0, x1, y0, y1)[0]


@jit
def transition_function_derivative(x, x0, x1, y0=0.0, y1=1.0):
    """Return derivative of the transition function

    This function simply returns the derivative of `transition_function` with respect to the `x`
    parameter.  The parameters to this function are identical to those of that function.

    Parameters
    ==========
    x: array_like
        One-dimensional monotonic array of floats.
    x0: float
        Value before which the output will equal `y0`.
    x1: float
        Value after which the output will equal `y1`.
    y0: float [defaults to 0.0]
        Value of the output before `x0`.
    y1: float [defaults to 1.0]
        Value of the output after `x1`.

    """
    transition_prime = np.zeros_like(x)
    ydiff = y1 - y0
    i = 0
    while x[i] <= x0:
        i += 1
    while x[i] < x1:
        tau = (x[i] - x0) / (x1 - x0)
        exponent = 1.0 / tau - 1.0 / (1.0 - tau)
        if exponent >= maxexp:
            transition_prime[i] = 0.0
        else:
            exponential = np.exp(1.0 / tau - 1.0 / (1.0 - tau))
            transition_prime[i] = (
                -ydiff
                * exponential
                * (-1.0 / tau ** 2 - 1.0 / (1.0 - tau) ** 2)
                * (1 / (x1 - x0))
                / (1.0 + exponential) ** 2
            )
        i += 1
    return transition_prime


@jit
def bump_function(x, x0, x1, x2, x3, y0=0.0, y12=1.0, y3=0.0):
    """Return a smooth bump function that is constant outside (x0, x3) and inside (x1, x2).

    This uses the standard C^infinity function with derivatives of compact support to transition
    between the the given values.  By default, this is a standard bump function that is 0 outside of
    (x0, x3), and is 1 inside (x1, x2), but the constant values can all be adjusted optionally.

    Parameters
    ==========
    x: array_like
        One-dimensional monotonic array of floats.
    x0: float
        Value before which the output will equal `y0`.
    x1, x2: float
        Values between which the output will equal `y12`.
    x3: float
        Value after which the output will equal `y3`.
    y0: float [defaults to 0.0]
        Value of the output before `x0`.
    y12: float [defaults to 1.0]
        Value of the output after `x1` but before `x2`.
    y3: float [defaults to 0.0]
        Value of the output after `x3`.

    """
    bump = np.empty_like(x)
    ydiff01 = y12 - y0
    ydiff23 = y3 - y12
    i = 0
    while x[i] <= x0:
        i += 1
    bump[:i] = y0
    while x[i] < x1:
        tau = (x[i] - x0) / (x1 - x0)
        exponent = 1.0 / tau - 1.0 / (1.0 - tau)
        if exponent >= maxexp:
            bump[i] = y0
        else:
            bump[i] = y0 + ydiff01 / (1.0 + np.exp(exponent))
        i += 1
    i1 = i
    while x[i] <= x2:
        i += 1
    bump[i1:i] = y12
    while x[i] < x3:
        tau = (x[i] - x2) / (x3 - x2)
        exponent = 1.0 / tau - 1.0 / (1.0 - tau)
        if exponent >= maxexp:
            bump[i] = y12
        else:
            bump[i] = y12 + ydiff23 / (1.0 + np.exp(exponent))
        i += 1
    bump[i:] = y3
    return bump


def transition_to_constant(f, t, t1, t2):
    """Smoothly transition from the function to a constant.

    This works (implicitly) by multiplying the derivative of `f` with the transition function, and
    then integrating.  Using integration by parts, this simplifies to multiplying `f` itself by the
    transition function, and then subtracting the integral of `f` times the derivative of the
    transition function.  This integral is effectively restricted to the region (t1, t2).  Note that
    the final value (after t2) will depend on the precise values of `t1` and `t2`, and the behavior
    of `f` in between.

    Parameters
    ==========
    f: array_like
        One-dimensional array corresponding to the following `t` parameter.
    t: array_like
        One-dimensional monotonic array of floats.
    t1: float
        Value before which the output will equal `f`.
    t2: float
        Value after which the output will be constant.

    """
    from quaternion import indefinite_integral

    transition, i1, i2 = transition_function(t, t1, t2, y0=1.0, y1=0.0, return_indices=True)
    transition_dot = transition_function_derivative(t, t1, t2, y0=1.0, y1=0.0)
    f_transitioned = f * transition
    f_transitioned[i1:i2] -= indefinite_integral(f[i1:i2] * transition_dot[i1:i2], t[i1:i2])
    f_transitioned[i2:] = f_transitioned[i2 - 1]
    return f_transitioned


@jit
def xor_timeseries(c):
    """XOR a time-series of data in place

    Assumes time varies along the first dimension of the input array, but any number of other
    dimensions are supported.

    This function leaves the first time step unchanged, but successive timesteps are the XOR from
    the preceding time step — storing only the bits that have changed.  This transformation is
    useful when storing the data because it allows for greater compression in many cases.

    Note that the direction in which this operation is done matters.  This function starts with the
    last time, changes that data in place, and proceeds to earlier times.  To undo this
    transformation, we need to start at early times and proceed to later times.

    The function `xor_timeseries_reverse` achieves the opposite transformation, recovering the
    original data with bit-for-bit accuracy.

    """
    u = c.view(np.uint64)
    for i in range(u.shape[0] - 1, 0, -1):
        u[i] = np.bitwise_xor(u[i - 1], u[i])
    return c


@jit
def xor_timeseries_reverse(c):
    """XOR a time-series of data in place

    This function reverses the effects of `xor_timeseries`.  See that function's docstring for
    details.

    """
    u = c.view(np.uint64)
    for i in range(1, u.shape[0]):
        u[i] = np.bitwise_xor(u[i - 1], u[i])
    return c


@jit
def fletcher32(data):
    """Compute the Fletcher-32 checksum of an array

    This checksum is very easy to implement from scratch and very fast.

    Note that it's not entirely clear that everyone agrees on the naming of
    these functions.  This version uses 16-bit input, 32-bit accumulators,
    block sizes of 360, and a modulus of 65_535.

    Parameters
    ==========
    data: ndarray
        This array can have any dtype, but must be able to be viewed as uint16.

    Returns
    =======
    checksum: uint32

    """
    data = data.reshape((data.size,)).view(np.uint16)
    size = data.size
    c0 = np.uint32(0)
    c1 = np.uint32(0)
    j = 0
    block_size = 360  # largest number of sums that can be performed without overflow
    while j < size:
        block_length = min(block_size, size - j)
        for i in range(block_length):
            c0 += data[j]
            c1 += c0
            j += 1
        c0 %= np.uint32(65535)
        c1 %= np.uint32(65535)
    return c1 << np.uint32(16) | c0


@functools.lru_cache()
def multishuffle(shuffle_widths, forward=True):
    """Construct functions to "multi-shuffle" data

    The standard "shuffle" algorithm (as found in HDF5, for example) takes an
    array of numbers and shuffles their bytes so that all bytes of a given
    significance are stored together — the first byte of all the numbers are
    stored contiguously, then the second byte of all the numbers, and so on.
    The motivation for this is that — with reasonably smooth data — bytes in
    the same byte position in sequential numbers are usually more related to
    each other than they are to other bytes within the same number, which means
    that shuffling results in better compression of the data.

    There is no reason that shuffling can only work byte-by-byte, however.
    There is also a "bitshuffle" algorithm, which works in the same way, but
    collecting bits rather than bytes.  More generally, we could vary the
    number of bits stored together as we move along the numbers.  For example,
    we could store the first 8 bits of each number, followed by the next 4 bits
    of each number, etc.  This is the "multi-shuffle" algorithm.

    With certain types of data, this can reduce the compressed data size
    significantly.  For example, with float data for which successive values
    have been XOR-ed, the sign bit will very rarely change, the next 11 bits
    (representing the exponent) and a few of the following bits (representing
    the highest-significance digits) will typically be highly correlated, while
    as we move to lower significance there will be less correlation.  Thus, we
    might shuffle the first 8 bits together, followed by the next 8, then the
    next 4, the next 4, the next 2, and so on — decreasing the shuffle width as
    we go.  The `shuffle_widths` input might look like [8, 8, 4, 4, 2, 2, 1, 1,
    1, 1, ...].

    There are also some cases where we see correlation *increasing* again at
    low significance.  For example, if a number results from cancellation — the
    subtraction of two numbers much larger than their difference — then its
    lower-significance bits will be 0.  If we then multiply that by some
    integer (e.g., for normalization), there may be some very correlated but
    nonzero pattern.  In either case, compression might improve if the values
    at the end of our shuffle_widths list increase.

    Parameters
    ==========
    shuffle_widths: list of integers
        These integers represent the number of bits in each piece of each
        number that is shuffled, starting from the highest significance, and
        proceeding to the lowest.  The sum of these numbers must be the total
        bit width of the numbers that will be given as input — which must
        currently be 8, 16, 32, or 64.  There is no restriction on the
        individual widths, but note that if they do not fit evenly into 8-bit
        bytes, the result is unlikely to compress well.
    forward: bool [defaults to True]
        If True, the returned function will shuffle data; if False, the
        returned function will reverse this process — unshuffle.

    Returns
    =======
    shuffle_func: numba JIT function
        This function takes just one parameter — the array to be shuffled — and
        returns the shuffled array.  Note that the input array *must* be flat
        (have just one dimension), and will be viewed as an array of unsigned
        integers of the input bit width.  This can affect the shape of the
        array and order of elements.  You should ensure that this process will
        result in an array of numbers in the order that you want.  For example,
        if you have a 2-d array of floats `a` that are more continuous along
        the first dimension, you might pass `np.ravel(a.view(np.uint64), 'F')`,
        where F represents Fortran order, which varies more quickly in the
        first dimension.

    """
    import numpy as np
    import numba as nb

    bit_width = np.sum(shuffle_widths, dtype=np.int64)  # uint64 casts to float under floor division...
    if bit_width not in [8, 16, 32, 64]:
        raise ValueError(f"Total bit width must be one of [8, 16, 32, 64], not {bit_width}")
    dtype = np.dtype(f"u{bit_width//8}")
    bit_width = dtype.type(bit_width)
    reversed_shuffle_widths = np.array(list(reversed(shuffle_widths)), dtype=dtype)
    one = dtype.type(1)

    if forward:

        def shuffle(a):
            a = a.view(dtype)
            if a.ndim != 1:
                raise ValueError(
                    "\nThis function only accepts flat arrays.  Make sure you flatten "
                    "(using ravel, reshape, or flatten)\n in a way that keeps your data"
                    "contiguous in the order you want."
                )
            b = np.zeros_like(a)
            b_array_bit = np.uint64(0)
            for i, shuffle_width in enumerate(reversed_shuffle_widths):
                mask_shift = np.sum(reversed_shuffle_widths[:i])
                mask = dtype.type(2 ** shuffle_width - 1)
                pieces_per_element = bit_width // shuffle_width
                for a_array_index in range(a.size):
                    b_array_index = b_array_bit // bit_width
                    b_element_bit = dtype.type(b_array_bit % bit_width)
                    masked = (a[a_array_index] >> mask_shift) & mask
                    b[b_array_index] += masked << b_element_bit
                    if b_element_bit + shuffle_width > bit_width:
                        b[b_array_index + one] += masked >> (bit_width - b_element_bit)
                    b_array_bit += shuffle_width
            return b

        return nb.jit(shuffle)

    else:
        # This function is almost the same as above, except for:
        # 1) swap a <-> b in input and output
        # 2) reverse the effect of the line in which b was set from a
        def unshuffle(b):
            b = b.view(dtype)
            if b.ndim != 1:
                raise ValueError(
                    "\nThis function only accepts flat arrays.  Make sure you flatten "
                    "(using ravel, reshape, or flatten)\n in a way that keeps your data"
                    "contiguous in the order you want."
                )
            a = np.zeros_like(b)
            b_array_bit = dtype.type(0)
            for i, shuffle_width in enumerate(reversed_shuffle_widths):
                mask_shift = np.sum(reversed_shuffle_widths[:i])
                mask = dtype.type(2 ** shuffle_width - 1)
                pieces_per_element = bit_width // shuffle_width
                for a_array_index in range(a.size):
                    b_array_index = b_array_bit // bit_width
                    b_element_bit = b_array_bit % bit_width
                    masked = (b[b_array_index] >> b_element_bit) & mask
                    a[a_array_index] += masked << mask_shift
                    if b_element_bit + shuffle_width > bit_width:
                        a[a_array_index] += (
                            (b[b_array_index + one] << (bit_width - b_element_bit)) & mask
                        ) << mask_shift
                    b_array_bit += shuffle_width
            return a

        return nb.jit(unshuffle)



@classmethod
def from_initial_values(cls, time, ell_max=8, sigma0=0.0, sigmadot0=0.0, sigmaddot0=0.0, psi2=0.0, psi1=0.0, psi0=0.0):
    """Construct Bondi data from sigma as a function of time and optional initial values

    The initial-value formulation for Bondi gauge is determined by these relations:

        ψ̇₀ = ðψ₁ + 3 σ ψ₂
        ψ̇₁ = ðψ₂ + 2 σ ψ₃
        ψ̇₂ = ðψ₃ + 1 σ ψ₄
        ψ₃ = -∂ðσ̄/∂u
        ψ₄ = -∂²σ̄/∂u²

    We also have a constraint on the initial value of ψ₂:

        Im[ψ₂] = -Im[ð²σ̄ + σ ∂σ̄/∂u]

    Given these expressions, and the value of sigma as a function of time and
    direction, we can find the values of sigma and all the Weyl components for all
    time -- the Bondi data.  This function implements that algorithm.

    Note that the interpretation of the input mode weights depends on the dimension
    of `sigma0`.  If it has dimension 0 or 1, the problem is integrated exactly,
    and the rest of the inputs are assumed to represent the corresponding values at
    time u=0.  If it has dimension 2, `sigma0` is interpreted as a function of
    time, and the rest of the problem is integrated numerically via splines, with
    `psi2`, `psi1`, and `psi0` representing the corresponding values at time
    u=`time[0]` (the first element of the input `time` array); `sigmadot0` and
    `sigmaddot0` are ignored in this case.

    Parameters
    ==========
    cls: class
        Class to construct the AsymptoticBondiData object.  This function will also
        be used as a classmethod in that class, in which case this parameter will
        be passed automatically by calling this function as a method.
    time: array_like
        Times at which to compute the Bondi data.  Must be 1-dimensional.
    ell_max: int
        Maximum ell value to be stored in the data
    sigma0: 0.0 or array_like [defaults to 0.0]
        This represents the value of sigma as a function of time and direction, or
        simply its initial value (see discussion above).  The input quantity must
        broadcast against an array of shape (n_times, LM_total_size(0, ell_max)).
    sigmadot0: 0.0 or array_like [defaults to 0.0]
        This represents the time-derivative of sigma as a function of time and
        direction.  Note that this is ignored if `sigma0` is a 2-dimensional array;
        instead, this quantity is computed by differentiating a spline of `sigma0`.
        This must be 0- or 1-dimensional, and is assumed to represent the
        derivative at time 0.  (Or, if `sigmaddot0` is 0, then the constant
        derivative at any time.)
    sigmaddot0: 0.0 or array_like [defaults to 0.0]
        Just like `sigmadot0`, except for the second derivative.  Note that this
        represents the second derivative only at time u=0; for all other times we
        expand as a Taylor series if this is used.
    psi2: 0.0 or array_like [defaults to 0.0]
        This represents the initial value of the psi2 field.  Its imaginary part is
        determined by the condition that the mass aspect must be real-valued, so
        any imaginary part in the input is discarded.  If array_like, this
        parameter must be 1-dimensional and have size LM_total_size(0, ell_max).
    psi1: 0.0 or array_like [defaults to 0.0]
        This represents the initial value of the psi1 field.  If array_like, this
        parameter must be 1-dimensional and have size LM_total_size(0, ell_max).
    psi0: 0.0 or array_like [defaults to 0.0]
        This represents the initial value of the psi0 field.  If array_like, this
        parameter must be 1-dimensional and have size LM_total_size(0, ell_max).

    """
    import numpy as np
    from .. import ModesTimeSeries

    def asany_atleast2d_complex(a):
        a = np.asanyarray(a) + 0j
        while np.ndim(a) < 2:
            a = a[np.newaxis, ...]
        return a

    psi2 = asany_atleast2d_complex(psi2)
    psi1 = asany_atleast2d_complex(psi1)
    psi0 = asany_atleast2d_complex(psi0)

    # Construct the empty container
    abd = cls(time, ell_max, multiplication_truncator=max)

    # Evaluate sigma and derivatives
    if np.ndim(sigma0) == 0 or np.ndim(sigma0) == 1:
        # Assume this is just the angular dependence, which will be taken as
        # constant in time.  If this is true, assumes sigmadot0 and sigmaddot0
        # are constants in time, and just integrates them.
        u = abd.time
        ð = lambda x: x.eth_GHP
        conjugate = lambda x: x.bar
        σ_0 = ModesTimeSeries(asany_atleast2d_complex(sigma0), u, spin_weight=2, multiplication_truncator=max)
        σ_1 = ModesTimeSeries(asany_atleast2d_complex(sigmadot0), u, spin_weight=2, multiplication_truncator=max)
        σ_2 = ModesTimeSeries(asany_atleast2d_complex(sigmaddot0 / 2), u, spin_weight=2, multiplication_truncator=max)
        abd.sigma = u * (u * σ_2 + σ_1) + σ_0
        # ψ₄ = -∂²σ̄/∂u²
        ψ4_0 = -2 * conjugate(σ_2)
        abd.psi4 = ψ4_0
        # ψ₃ = -ð ∂σ̄/∂u
        ψ3_0 = -ð(conjugate(σ_1))
        ψ3_1 = -2 * ð(conjugate(σ_2))
        abd.psi3 = u * ψ3_1 + ψ3_0
        # ψ₂ = ∫ (ðψ₃ + σψ₄) du
        ψ2_0 = (
            ModesTimeSeries(psi2, u, spin_weight=0, multiplication_truncator=max).real
            - 1j * (σ_0.bar.eth_GHP.eth_GHP + σ_0 * σ_1.bar).imag
        )
        ψ2_1 = σ_0 * ψ4_0 + ð(ψ3_0)
        ψ2_2 = (σ_1 * ψ4_0 + ð(ψ3_1)) / 2
        ψ2_3 = (1 / 3) * σ_2 * ψ4_0
        abd.psi2 = u * (u * (u * ψ2_3 + ψ2_2) + ψ2_1) + ψ2_0
        # ψ₁ = ∫ (ðψ₂ + 2σψ₃) du
        ψ1_0 = ModesTimeSeries(psi1, u, spin_weight=1, multiplication_truncator=max)
        ψ1_1 = 2 * σ_0 * ψ3_0 + ð(ψ2_0)
        ψ1_2 = σ_0 * ψ3_1 + σ_1 * ψ3_0 + ð(ψ2_1) / 2
        ψ1_3 = (2 * σ_1 * ψ3_1 + 2 * σ_2 * ψ3_0 + ð(ψ2_2)) / 3
        ψ1_4 = (2 * σ_2 * ψ3_1 + ð(ψ2_3)) / 4
        abd.psi1 = u * (u * (u * (u * ψ1_4 + ψ1_3) + ψ1_2) + ψ1_1) + ψ1_0
        # ψ₀ = ∫ (ðψ₁ + 3σψ₂) du
        ψ0_0 = ModesTimeSeries(psi0, u, spin_weight=2, multiplication_truncator=max)
        ψ0_1 = 3 * σ_0 * ψ2_0 + ð(ψ1_0)
        ψ0_2 = (3 * σ_0 * ψ2_1 + 3 * σ_1 * ψ2_0 + ð(ψ1_1)) / 2
        ψ0_3 = σ_0 * ψ2_2 + σ_1 * ψ2_1 + σ_2 * ψ2_0 + ð(ψ1_2) / 3
        ψ0_4 = (3 * σ_0 * ψ2_3 + 3 * σ_1 * ψ2_2 + 3 * σ_2 * ψ2_1 + ð(ψ1_3)) / 4
        ψ0_5 = (3 * σ_1 * ψ2_3 + 3 * σ_2 * ψ2_2 + ð(ψ1_4)) / 5
        ψ0_6 = σ_2 * ψ2_3 / 2
        abd.psi0 = u * (u * (u * (u * (u * (u * ψ0_6 + ψ0_5) + ψ0_4) + ψ0_3) + ψ0_2) + ψ0_1) + ψ0_0
    elif np.ndim(sigma0) == 2:
        # Assume this gives complete data, as a function of time and angle.
        # If this is true, ignore sigmadot0 and sigmaddot0.
        def adjust_psi2_imaginary_part(psi2, abd):
            # Adjust the initial value of psi2 to satisfy the mass-aspect condition
            sigma_initial = abd.sigma[..., 0, :]
            sigma_bar_dot_initial = abd.sigma.bar.dot[..., 0, :]
            return (
                ModesTimeSeries(psi2, abd.time, spin_weight=0).real
                - 1j * (sigma_initial.bar.eth_GHP.eth_GHP + sigma_initial * sigma_bar_dot_initial).imag
            )

        abd.sigma = sigma0
        # ψ₄ = -∂²σ̄/∂u²
        abd.psi4 = -abd.sigma.ddot.bar
        # ψ₃ = -ð ∂σ̄/∂u
        abd.psi3 = -abd.sigma.dot.bar.eth_GHP
        # ψ₂ = ∫ (ðψ₃ + σψ₄) du
        abd.psi2 = (abd.psi3.eth_GHP + abd.sigma * abd.psi4).int + adjust_psi2_imaginary_part(psi2, abd)
        # ψ₁ = ∫ ðψ₂ + σψ₃ du
        abd.psi1 = (abd.psi2.eth_GHP + 2 * abd.sigma * abd.psi3).int + psi1
        # ψ₀ = ∫ ðψ₁ + σψ₂ dt
        abd.psi0 = (abd.psi1.eth_GHP + 3 * abd.sigma * abd.psi2).int + psi0
    else:
        raise ValueError(f"Input `sigma0` must have 1 or 2 dimensions; it has {np.ndim(sigma0)}")

    return abd



import numpy as np
from numpy.polynomial.polynomial import polyfit

mode_regex = r"""Y_l(?P<L>[0-9]+)_m(?P<M>[-+0-9]+)\.dat"""


def pick_Ch_mass(filename="Horizons.h5"):
    """Deduce the best Christodoulou mass by finding the statistical "mode" (after binning)."""
    from h5py import File
    from os.path import isdir
    from numpy import histogram

    if isdir(filename):
        filename = filename + "Horizons.h5"
    try:
        f = File(filename, "r")
    except OSError:
        print(f"pick_Ch_mass could not open the file '{filename}'")
        raise
    ChMass = f["AhA.dir/ChristodoulouMass.dat"][:, 1] + f["AhB.dir/ChristodoulouMass.dat"][:, 1]
    f.close()
    hist, bins = histogram(ChMass, bins=len(ChMass))
    return bins[hist.argmax()]


def monotonic_indices(T, MinTimeStep=1.0e-3):
    """Given an array of times, return the indices that make the array strictly monotonic."""
    from numpy import delete

    Ind = range(len(T))
    Size = len(Ind)
    i = 1
    while i < Size:
        if T[Ind[i]] <= T[Ind[i - 1]] + MinTimeStep:
            j = 0
            while T[Ind[j]] + MinTimeStep < T[Ind[i]]:
                j += 1
            # erase data from j (inclusive) to i (exclusive)
            Ind = delete(Ind, range(j, i))
            Size = len(Ind)
            i = j - 1
        i += 1
    return Ind


def intersection(t1, t2, min_step=None, min_time=None, max_time=None):
    """Return the intersection of two time sequences.

    The time step at each point is the minimum of the time steps in
    t1 and t2 at that instant, or min_step, whichever is greater.
    The output starts at the earliest moment common to t1 and t2, or
    min_time, whichever is greater.

    Parameters
    ----------
    t1: 1-d float array
    t2: 1-d float array
    min_step: float
    min_time: float
    max_time: float

    Returns
    -------
    1-d float array

    """
    import numpy as np

    t1 = np.asarray(t1)
    t2 = np.asarray(t2)
    if t1.size == 0:
        raise ValueError("t1 is empty.  Assuming this is not desired.")
    if t2.size == 0:
        raise ValueError("t2 is empty.  Assuming this is not desired.")
    t = np.empty(t1.size + t2.size)
    min1 = t1[0]
    min2 = t2[0]
    if min_time is None:
        mint = max(min1, min2)
    else:
        mint = max(max(min1, min2), min_time)
    max1 = t1[-1]
    max2 = t2[-1]
    if max_time is None:
        maxt = min(max1, max2)
    else:
        maxt = min(min(max1, max2), max_time)
    if mint > max1 or mint > max2:
        message = "Empty intersection in t1=[{0}, ..., {1}], t2=[{2}, ..., {3}] " + "with min_time={4}"
        raise ValueError(message.format(min1, max1, min2, max2, min_time))
    if maxt < min1 or maxt < min2:
        message = "Empty intersection in t1=[{0}, ..., {1}], t2=[{2}, ..., {3}] " + "with max_time={4}"
        raise ValueError(message.format(min1, max1, min2, max2, max_time))
    if min_step is None:
        min_step = min(np.min(np.diff(t1)), np.min(np.diff(t2)))
    t[0] = mint
    I = 0
    I1 = 0
    I2 = 0
    size1 = t1.size
    size2 = t2.size
    while t[I] < maxt:
        # adjust I1 to ensure that t[I] is in the interval ( t1[I1-1], t1[I1] ]
        if t[I] < min1 or t[I] > max1:
            # if t[I] is less than the smallest t1, or greater than the largest t1, I1=0
            I1 = 0
        else:
            I1 = max(I1, 1)
            while t[I] > t1[I1] and I1 < size1:
                I1 += 1
        # adjust I2 to ensure that t[I] is in the interval ( t2[I2-1], t2[I2] ]
        if t[I] < min2 or t[I] > max2:
            # if t[I] is less than the smallest t2, or greater than the largest t2, I2=0
            I2 = 0
        else:
            I2 = max(I2, 1)
            while t[I] > t2[I2] and I2 < size2:
                I2 += 1
        t[I + 1] = t[I] + max(min(t1[I1] - t1[I1 - 1], t2[I2] - t2[I2 - 1]), min_step)
        I += 1
        if t[I] > maxt:
            break
    return t[:I]  # only take the relevant part of the reserved vector


def validate_single_waveform(h5file, filename, WaveformName, ExpectedNModes, ExpectedNTimes, LModes):
    # from sys import stderr
    from re import compile as re_compile

    CompiledModeRegex = re_compile(mode_regex)
    Valid = True
    # Check ArealRadius
    if not h5file[WaveformName + "/ArealRadius.dat"].shape == (ExpectedNTimes, 2):
        Valid = False
        print(
            "{}:{}/ArealRadius.dat\n\tGot shape {}; expected ({}, 2)".format(
                filename,
                WaveformName,
                h5file[WaveformName + "/ArealRadius.dat"].shape,
                ExpectedNTimes,
            )
        )
    # Check AverageLapse
    if not h5file[WaveformName + "/AverageLapse.dat"].shape == (ExpectedNTimes, 2):
        Valid = False
        print(
            "{}:{}/AverageLapse.dat\n\tGot shape {}; expected ({}, 2)".format(
                filename,
                WaveformName,
                h5file[WaveformName + "/AverageLapse.dat"].shape,
                ExpectedNTimes,
            )
        )
    # Check Y_l*_m*.dat
    NModes = len(
        [
            True
            for dataset in list(h5file[WaveformName])
            for m in [CompiledModeRegex.search(dataset)]
            if m and int(m.group("L")) in LModes
        ]
    )
    if not NModes == ExpectedNModes:
        Valid = False
        print(
            "{}:{}/{}\n\tGot {} modes; expected {}".format(filename, WaveformName, mode_regex, NModes, ExpectedNModes)
        )
    for dataset in list(h5file[WaveformName]):
        if CompiledModeRegex.search(dataset):
            if not h5file[WaveformName + "/" + dataset].shape == (ExpectedNTimes, 3):
                Valid = False
                (
                    "{}:{}/{}\n\tGot shape {}; expected ({}, 3)".format(
                        filename,
                        WaveformName,
                        dataset,
                        h5file[WaveformName + "/" + dataset].shape,
                        ExpectedNTimes,
                    )
                )
    return Valid


def validate_group_of_waveforms(h5file, filename, WaveformNames):
    from re import compile as re_compile
    import scri

    DataType = datatype_from_filename(filename)
    # Set the correct LModes based on the spin-weight
    if DataType == scri.psi3 or DataType == scri.psi1:
        LModes = range(1, 200)
    elif DataType == scri.psi2:
        LModes = range(0, 200)
    else:
        LModes = range(2, 200)

    ExpectedNTimes = h5file[WaveformNames[0] + "/ArealRadius.dat"].shape[0]
    ExpectedNModes = len(
        [
            True
            for dataset in list(h5file[WaveformNames[0]])
            for m in [re_compile(mode_regex).search(dataset)]
            if m and int(m.group("L")) in LModes
        ]
    )
    Valid = True
    FailedWaveforms = []
    for WaveformName in WaveformNames:
        if not validate_single_waveform(h5file, filename, WaveformName, ExpectedNModes, ExpectedNTimes, LModes):
            Valid = False
            FailedWaveforms.append(WaveformName)
    if not Valid:
        # from sys import stderr
        print("In '{}', the following waveforms are not valid:\n\t{}".format(filename, "\n\t".join(FailedWaveforms)))
    return Valid

def datatype_from_filename(filename):
    from os.path import basename
    import scri

    DataType = basename(filename).partition("_")[0]
    if "hdot" in DataType.lower():
        DataType = scri.hdot
    elif "h" in DataType.lower():
        DataType = scri.h
    elif "psi4" in DataType.lower():
        DataType = scri.psi4
    elif "psi3" in DataType.lower():
        DataType = scri.psi3
    elif "psi2" in DataType.lower():
        DataType = scri.psi2
    elif "psi1" in DataType.lower():
        DataType = scri.psi1
    elif "psi0" in DataType.lower():
        DataType = scri.psi0
    else:
        DataType = scri.UnknownDataType
        message = (
            "The file '{0}' does not contain a recognizable " +
            "description "+
            "of the data type ('h', 'psi4', 'psi3', 'psi2', "+
            "'psi1', 'psi0')."
        )
        raise ValueError(message.format(filename))
    return DataType

def read_finite_radius_waveform_nrar(filename, WaveformName):

    from h5py import File
    from numpy import array
    import scri
    import os

    # Open the file twice.
    # The first time is for the auxiliary quantities.
    with File(filename,"r") as f:
        W = f[WaveformName]
        
        # Read the time, account for repeated indices
        T = W["AverageLapse.dat"][:, 0]
        Indices = monotonic_indices(T)
        T = T[Indices]

        # Read the other auxiliary quantities
        Radii = array(W["ArealRadius.dat"])[Indices, 1]
        AverageLapse = array(W["AverageLapse.dat"])[Indices, 1]
        CoordRadius = W["CoordRadius.dat"][0, 1]
        InitialAdmEnergy = W["InitialAdmEnergy.dat"][0, 1]

    # The second time we read the file is for the waveform
    waveform = scri.SpEC.file_io.read_from_h5(
        os.path.join(filename,WaveformName),
        frameType=scri.Inertial,
        dataType=datatype_from_filename(filename),
        r_is_scaled_out=True,
        m_is_scaled_out=False, # For now. We will change this later.
    )
    
    return waveform,T,Indices,Radii,AverageLapse,CoordRadius,InitialAdmEnergy

def read_finite_radius_waveform_rpxmb_or_rpdmb(filename,
                                               groupname, WaveformName):
    """This is just a worker function defined for read_finite_radius_data, 
       below, reading a single waveform from an h5 file of many waveforms. 
       You probably don't need to call this directly.
    """

    from h5py import File
    import os
    import scri
    import sxs
    import json
    from numpy import array

    # Open the file twice.
    # The first time reads the waveform.
    rpdmb_formats=["rotating_paired_diff_multishuffle_bzip2", "rpdmb", "RPDMB"]
    rpxmb_formats=["rotating_paired_xor_multishuffle_bzip2", "rpxmb", "rpxm",
                   "RPXMB", "RPXM"]
    json_path = filename.replace(".h5",".json")
    read_rpxmb = False
    read_rpdmb = False
    if os.path.exists(json_path):
        with open(json_path) as f:
            full_group_name = groupname + '/' + WaveformName
            json_data = json.load(f)[full_group_name]
        sxs_format = json_data.get("sxs_format","")
        if sxs_format in rpdmb_formats:
            read_rpdmb=True
        elif sxs_format in rpdmb_formats:
            read_rpxmb=True

    # Note that groupname begins with a '/' so
    # os.path.join(filename,groupname,WaveformName) does not work.
    if read_rpxmb:
        waveform=scri.rpxmb.load(filename+groupname+"/"+
                                 WaveformName)[0].to_inertial_frame()
    elif read_rpdmb:
        waveform=scri.WaveformModes.from_sxs(
            sxs.rpdmb.load(filename+groupname+"/"+WaveformName))
    else:
        raise ValueError("Cannot find rpxmb/rpdmb format string "
                         " in {}, group {}".format(json_path,groupname))
    T = waveform.t

    # The second time we read the file is for the auxiliary quantities.
    with File(filename,"r") as f:
        W = f[groupname][WaveformName]

        # Account for repeated indices in time.
        Indices = monotonic_indices(T)
        T = T[Indices]

        # Read the other auxiliary quantities
        Radii = array(W["ArealRadius.dat"])[Indices]
        AverageLapse = array(W["AverageLapse.dat"])[Indices]
        CoordRadius = W["CoordRadius.dat"][1]
        InitialAdmEnergy = W["InitialAdmEnergy.dat"][1]

    return waveform,T,Indices,Radii,AverageLapse,CoordRadius,InitialAdmEnergy

def read_finite_radius_waveform(filename, groupname, WaveformName, ChMass):
    from scipy.integrate import cumtrapz as integrate
    from numpy import log, sqrt
    import scri

    if groupname is None:
       waveform,T,Indices,Radii,AverageLapse,CoordRadius,InitialAdmEnergy = \
           read_finite_radius_waveform_nrar(filename,WaveformName)
    else:
       waveform,T,Indices,Radii,AverageLapse,CoordRadius,InitialAdmEnergy = \
           read_finite_radius_waveform_rpxmb_or_rpdmb(filename,
                                                      groupname,WaveformName)

    # Rescale and offset the time array so that the time array is
    # approximately the tortoise coordinate.
    T[1:] = integrate(AverageLapse / sqrt(((-2.0 * InitialAdmEnergy) / Radii)\
                                          + 1.0), T) + T[0]
    T -= Radii + (2.0 * InitialAdmEnergy) \
        * log((Radii / (2.0 * InitialAdmEnergy)) - 1.0)
   
    # Now determine the scaling with mass.
    if waveform.dataType == scri.h:
        UnitScaleFactor = 1.0 / ChMass
        RadiusRatioExp = 1.0
    elif waveform.dataType == scri.hdot:
        UnitScaleFactor = 1.0
        RadiusRatioExp = 1.0
    elif waveform.dataType == scri.psi4:
        UnitScaleFactor = ChMass
        RadiusRatioExp = 1.0
    elif waveform.dataType == scri.psi3:
        UnitScaleFactor = 1.0
        RadiusRatioExp = 2.0
    elif waveform.dataType == scri.psi2:
        UnitScaleFactor = 1.0 / ChMass
        RadiusRatioExp = 3.0
    elif waveform.dataType == scri.psi1:
        UnitScaleFactor = 1.0 / ChMass ** 2
        RadiusRatioExp = 4.0
    elif waveform.dataType == scri.psi0:
        UnitScaleFactor = 1.0 / ChMass ** 3
        RadiusRatioExp = 5.0
    else:
        raise ValueError(f'DataType "{waveform.dataType}" is unknown.')

    # The read_finite_radius_waveform_* functions removed nonmonotonic times
    # from T but did not do the same for the waveform data. So we do that here.
    waveform.data = waveform.data[Indices,:]

    # We also need to remove nonmonontonic times from the frame, if the
    # frame has the appropriate shape.
    if waveform.frame.shape[0] > 1:
        waveform.frame = waveform.frame[Indices]

    # Rescale the times and the data
    RadiusRatio = (Radii / CoordRadius) ** RadiusRatioExp
    waveform.t = T/ChMass
    for m,_ in enumerate(waveform.LM):
        waveform.data[:,m] *= RadiusRatio * UnitScaleFactor
    waveform.m_is_scaled_out = True

    # Add the history information
    history =  (
        """# extrapolation.read_finite_radius_waveform""" +
        """({0}, {1}, {2}, {3})"""
    )
    history = history.format(
        filename,
        groupname,
        WaveformName,
        ChMass,
    )
    waveform.history=[history]

    return waveform, Radii/ChMass
        
def read_finite_radius_data(ChMass=0.0,
                            filename="rh_FiniteRadii_CodeUnits.h5",
                            CoordRadii=[]):
    """Read data at various radii, and offset by tortoise coordinate."""

    if ChMass == 0.0:
        raise ValueError("ChMass=0.0 is not a valid input value.")

    from sys import stdout, stderr
    from os.path import basename
    from h5py import File
    from re import compile as re_compile
    import scri

    YLMRegex = re_compile(mode_regex)
    
    # If 'filename' is of the form "h5_file_name.h5/groupname" then we have an
    # RPXMB file, and groupname identifies the quantity we are extrapolating.
    # Otherwise, we have a NRAR finite-radius file and the filename identifies
    # the quantity we are extrapolating.
    groupname = None
    if ".h5" in filename and not filename.endswith(".h5"):
        filename, groupname = filename.split(".h5")
        filename = filename + ".h5"
        if groupname == "/":
            groupname = None

    try:
        f = File(filename, "r")
    except OSError:
        print(f"read_finite_radius_data could not open the file '{filename}'")
        raise
    try:
        # Get list of waveforms we'll be using
        if groupname is None:
            WaveformNames = list(f)
            # VersionHist.ver is not one of the WaveformNames.
            if "VersionHist.ver" in WaveformNames:
                WaveformNames.remove("VersionHist.ver")
        else:
            WaveformNames = list(f[groupname])
        if not CoordRadii:
            # If the list of Radii is empty, figure out what they are
            CoordRadii = [
                m.group("r") for Name in WaveformNames for m in [re_compile(r"""R(?P<r>.*?)\.dir""").search(Name)] if m
            ]
        else:
            # Pare down the WaveformNames list appropriately
            if type(CoordRadii[0]) == int:
                WaveformNames = [WaveformNames[i] for i in CoordRadii]
                CoordRadii = [
                    m.group("r") for Name in CoordRadii
                    for m in
                    [ re_compile(r"""R(?P<r>.*?)\.dir""").search(Name)] if m ]
            else:
                WaveformNames = [
                    Name for Name in WaveformNames for Radius in
                    CoordRadii for m in [re_compile(Radius).search(Name)] if m]
        NWaveforms = len(WaveformNames)
        
        # Check input data for NRAR format
        if groupname is None:
            if not validate_group_of_waveforms(f, filename, WaveformNames):
                raise ValueError(f"Bad input waveforms in {filename}.")
            stdout.write(f"{filename} passed the data-integrity tests.\n")
            stdout.flush()
        Ws = [scri.WaveformModes() for i in range(NWaveforms)]
        Radii = [None] * NWaveforms

    finally:
        f.close()

    PrintedLine = ""
    for n in range(NWaveforms):
        if n == NWaveforms - 1:
            WaveformNameString = WaveformNames[n] + "\n"
        else:
            WaveformNameString = WaveformNames[n] + ", "
        if len(PrintedLine + WaveformNameString) > 100:
            stdout.write("\n" + WaveformNameString)
            stdout.flush()
            PrintedLine = WaveformNameString
        else:
            stdout.write(WaveformNameString)
            stdout.flush()
            PrintedLine += WaveformNameString
        Ws[n], Radii[n] = read_finite_radius_waveform(filename,groupname,
                                                      WaveformNames[n],
                                                      ChMass)
    return Ws, Radii, CoordRadii

def set_common_time(Ws, Radii, MinTimeStep=0.005, EarliestTime=-3e300, LatestTime=3e300):
    """Interpolate Waveforms and radius data to a common set of times."""
    from scipy import interpolate

    NWaveforms = len(Radii)
    TLimits = [EarliestTime, LatestTime]
    # Get the new time data before any interpolations
    T = intersection(TLimits, Ws[0].t, MinTimeStep, EarliestTime, LatestTime)
    for i_W in range(1, NWaveforms):
        T = intersection(T, Ws[i_W].t)
    # Interpolate Radii and then Ws (in that order!)
    for i_W in range(NWaveforms):
        Radii[i_W] = interpolate.InterpolatedUnivariateSpline(Ws[i_W].t, Radii[i_W])(T)
        Ws[i_W] = Ws[i_W].interpolate(T)
    return


def extrapolate(**kwargs):
    """Perform extrapolations from finite-radius data.

    See arXiv:2010.15200 for specific details.

    Parameters
    ----------
    InputDirectory : str, (Default: './')
        Where to find the input data.  Can be relative or absolute.

    OutputDirectory : str, (Default: './')
        This directory will be made if it does not exist.  If this is
        passed but is the empty string, no files of any kind will be
        output.

    DataFile : str, (Default: 'rh_FiniteRadii_CodeUnits.h5')
        Input file holding the data from all the radii.

    ChMass : float, (Default: 0.0)
        Christodoulou mass in the same units as the rest of the
        data.  All the data will be rescaled into units such that
        this is one.  If this is zero, the Christodoulou mass will
        be extracted automatically from the horizons file below.

    HorizonsFile : str, (Default: 'Horizons.h5')
        File name to read for horizon data (if ChMass is 0.0).

    CoordRadii : list of str or list of int, (Default: [])
        List of strings containing the radii to use, or of (integer)
        indices of the list of waveform names.  If this is a list of
        indices, the order is just the order output by the command
        `list(h5py.File(DataFile))` which *should* be the same as
        `h5ls`.  If the list is empty, all radii that can be found
        are used.

    ExtrapolationOrders : list of int, (Default: [-1, 2, 3, 4, 5, 6])
        Negative numbers correspond to extracted data, counting down
        from the outermost extraction radius (which is -1).

    UseOmega : bool, (Default: False)
        Whether or not to extrapolate as a function of lambda/r =
        1/(r*m*omega), where omega is the instantaneous angular
        frequency of rotation.  If this is True, the extrapolation
        will usually not converge for high N; if this is False, SVD
        will generally cause the convergence to appear to fall to
        roundoff, though the accuracy presumably is not so great.

    OutputFrame : {scri.Inertial, scri.Corotating}
        Transform to this frame before comparison and output.

    ExtrapolatedFiles : str, (Default: 'Extrapolated_N{N}.h5')
    DifferenceFiles   : str, (Default: 'ExtrapConvergence_N{N}-N{Nm1}.h5')
        These are python-formatted output file names, where the
        extrapolation order N is substituted for '{N}', and the
        previous extrapolation order is substituted for '{Nm1}'.
        The data-type inferred from the DataFile name is prepended.
        If DifferenceFiles is empty, the corresponding file is not
        output.

    UseStupidNRARFormat : bool, (Default: False)
        If True (and `ExtrapolatedFiles` does not end in '.dat'),
        then the h5 output format will be that stupid, old
        NRAR/NINJA format that doesn't convey enough information,
        is slow, and uses 33% more space than it needs to.  But you
        know, if you're into that kind of thing, whatever.  Who am
        I to judge?

    PlotFormat : str, (Default: '')
        The format of output plots.  This can be the empty string,
        in which case no plotting is done.  Or, these can be any of
        the formats supported by your installation of matplotlib.

    MinTimeStep : float, (Default: 0.005)
        The smallest allowed time step in the output data.

    EarliestTime : float, (Default: -3.0e300)
        The earliest time in the output data.  For values less than
        0, some of the data corresponds to times when only junk
        radiation is present.

    LatestTime : float, (Default: 3.0e300)
        The latest time in the output data.

    AlignmentTime : float, (Default: None)
        The time at which to align the Waveform with the dominant
        eigenvector of <LL>.  If the input value is `None` or is
        outside of the input data, it will be reset to the midpoint
        of the waveform: (W_outer.T(0)+W_outer.T(-1))/2

    NoiseFloor : float, (Default: None)
        ONLY USED FOR Psi1 AND Psi0 WAVEFORMS.
        The effective noise floor of the SpEC simulation.
        If Psi1 or Psi0 (NOT scaled by radius) falls beneath
        this value for certain extraction radii, then those
        radii are not included in the extrapolation. The value
        1.0e-9 seems to work well for a few BBH systems.

    """

    # Basic imports
    from os import makedirs, remove
    from os.path import exists, basename, dirname
    from sys import stdout, stderr
    from textwrap import dedent
    from numpy import sqrt, abs, fmod, pi, transpose, array
    from scipy.interpolate import splev, splrep
    import scri
    from scri import Inertial, Corotating, WaveformModes

    # Process keyword arguments
    InputDirectory = kwargs.pop("InputDirectory", "./")
    OutputDirectory = kwargs.pop("OutputDirectory", "./")
    DataFile = kwargs.pop("DataFile", "rh_FiniteRadii_CodeUnits.h5")
    ChMass = kwargs.pop("ChMass", 0.0)
    HorizonsFile = kwargs.pop("HorizonsFile", "Horizons.h5")
    CoordRadiiKwarg = kwargs.get("CoordRadii", None)
    CoordRadii = kwargs.pop("CoordRadii", [])
    ExtrapolationOrders = kwargs.pop("ExtrapolationOrders", [-1, 2, 3, 4, 5, 6])
    UseOmega = kwargs.pop("UseOmega", False)
    OutputFrame = kwargs.pop("OutputFrame", Inertial)
    ExtrapolatedFiles = kwargs.pop("ExtrapolatedFiles", "Extrapolated_N{N}.h5")
    DifferenceFiles = kwargs.pop("DifferenceFiles", "ExtrapConvergence_N{N}-N{Nm1}.h5")
    UseStupidNRARFormat = kwargs.pop("UseStupidNRARFormat", False)
    PlotFormat = kwargs.pop("PlotFormat", "")
    MinTimeStep = kwargs.pop("MinTimeStep", 0.005)
    EarliestTime = kwargs.pop("EarliestTime", -3.0e300)
    LatestTime = kwargs.pop("LatestTime", 3.0e300)
    AlignmentTime = kwargs.pop("AlignmentTime", None)
    NoiseFloor = kwargs.pop("NoiseFloor", None)
    return_finite_radius_waveforms = kwargs.pop("return_finite_radius_waveforms", False)
    if len(kwargs) > 0:
        raise ValueError(f"Unknown arguments to `extrapolate`: kwargs={kwargs}")

    # Polish up the input arguments
    if not InputDirectory.endswith("/"):
        InputDirectory += "/"
    if OutputDirectory and not OutputDirectory.endswith("/"):
        OutputDirectory += "/"
    if not exists(HorizonsFile):
        HorizonsFile = InputDirectory + HorizonsFile
    if not exists(DataFile):
        DataFile = InputDirectory + DataFile
    if ChMass == 0.0:
        print("WARNING: ChMass is being automatically determined from the data, " + "rather than metadata.txt.")
        ChMass = pick_Ch_mass(HorizonsFile)

    # AlignmentTime is reset properly once the data are read in, if necessary.
    # The reasonableness of ExtrapolationOrder is checked below.

    # Don't bother loading plotting modules unless we're plotting
    if PlotFormat:
        import matplotlib as mpl

        mpl.use("Agg")
        import matplotlib.pyplot as plt

        mpl.rcParams["axes.prop_cycle"] = mpl.cycler(
            "color", ["#000000", "#cc79a7", "#d55e00", "#0072b2", "#f0e442", "#56b4e9", "#e69f00", "#2b9f78",],
        )
        figabs = plt.figure(0)
        figarg = plt.figure(1)
        fignorm = plt.figure(2)

    # Read in the Waveforms
    print(f"Reading Waveforms from {DataFile}...")
    stdout.flush()
    Ws, Radii, CoordRadii = read_finite_radius_data(
        ChMass=ChMass, filename=DataFile, CoordRadii=CoordRadii)

    Radii_shape = (len(Radii), len(Radii[0]))

    # Make sure there are enough radii to do the requested extrapolations
    if (len(Ws) <= max(ExtrapolationOrders)) and (max(ExtrapolationOrders) > -1):
        raise ValueError(
            "Not enough data sets ({}) for max extrapolation order (N={}).".format(len(Ws), max(ExtrapolationOrders))
        )
    if -len(Ws) > min(ExtrapolationOrders):
        raise ValueError(
            "Not enough data sets ({}) for min extrapolation order (N={}).".format(len(Ws), min(ExtrapolationOrders))
        )

    # Figure out which is the outermost data
    SortedRadiiIndices = sorted(range(len(CoordRadii)), key=lambda k: float(CoordRadii[k]))
    i_outer = SortedRadiiIndices[-1]

    # Interpolate to common times
    print("Interpolating to common times...")
    stdout.flush()
    set_common_time(Ws, Radii, MinTimeStep, EarliestTime, LatestTime)
    W_outer = Ws[i_outer]

    # If the AlignmentTime is not set properly, set it to the default
    if (not AlignmentTime) or AlignmentTime < W_outer.t[0] or AlignmentTime >= W_outer.t[-1]:
        AlignmentTime = (W_outer.t[0] + W_outer.t[-1]) / 2.0

    # Print the input arguments neatly for the history
    InputArguments = """\
        # Extrapolation input arguments:
        D = {{}}
        D['InputDirectory'] = {InputDirectory}
        D['OutputDirectory'] = {OutputDirectory}
        D['DataFile'] = {DataFile}
        D['ChMass'] = {ChMass}
        D['HorizonsFile'] = {HorizonsFile}
        D['CoordRadii'] = {CoordRadii}
        D['ExtrapolationOrders'] = {ExtrapolationOrders}
        D['UseOmega'] = {UseOmega}
        D['OutputFrame'] = {OutputFrame}
        D['ExtrapolatedFiles'] = {ExtrapolatedFiles}
        D['DifferenceFiles'] = {DifferenceFiles}
        D['UseStupidNRARFormat'] = {UseStupidNRARFormat}
        D['PlotFormat'] = {PlotFormat}
        D['MinTimeStep'] = {MinTimeStep}
        D['EarliestTime'] = {EarliestTime}
        D['LatestTime'] = {LatestTime}
        D['AlignmentTime'] = {AlignmentTime}
        D['NoiseFloor'] = {NoiseFloor}
        # End Extrapolation input arguments
        """.format(
        InputDirectory=InputDirectory,
        OutputDirectory=OutputDirectory,
        DataFile=DataFile,
        ChMass=ChMass,
        HorizonsFile=HorizonsFile,
        CoordRadii=CoordRadii,
        ExtrapolationOrders=ExtrapolationOrders,
        UseOmega=UseOmega,
        OutputFrame=OutputFrame,
        ExtrapolatedFiles=ExtrapolatedFiles,
        DifferenceFiles=DifferenceFiles,
        UseStupidNRARFormat=UseStupidNRARFormat,
        PlotFormat=PlotFormat,
        MinTimeStep=MinTimeStep,
        EarliestTime=EarliestTime,
        LatestTime=LatestTime,
        AlignmentTime=AlignmentTime,
        NoiseFloor=NoiseFloor,
    )
    InputArguments = dedent(InputArguments)

    # If required, figure out the orbital frequencies
    if UseOmega:
        Omegas = [sqrt(sum([c ** 2 for c in o])) for o in W_outer.AngularVelocityVectorRelativeToInertial([2])]
    else:
        Omegas = []

    # Transform W_outer into its smoothed corotating frame, and align modes with
    # frame at given instant
    stdout.write("Rotating into common (outer) frame...\n")
    stdout.flush()
    if W_outer.frameType != Inertial:
        raise ValueError("Extrapolation assumes that the input data are in the inertial frame")
    print("Using alignment region (0.1, 0.8)")
    W_outer.to_corotating_frame(z_alignment_region=(0.1, 0.8))
    # W_outer.to_corotating_frame()
    # W_outer.align_decomposition_frame_to_modes(AlignmentTime)

    # Transform everyone else into the same frame
    for i in SortedRadiiIndices[:-1]:
        Ws[i].rotate_decomposition_basis(W_outer.frame)
        Ws[i].frameType = Corotating

    # Remove old h5 file if necessary
    if not ExtrapolatedFiles.endswith(".dat") and UseStupidNRARFormat:
        h5Index = ExtrapolatedFiles.find(".h5/")
        if h5Index > 0:
            if exists(ExtrapolatedFiles[: h5Index + 3]):
                remove(ExtrapolatedFiles[: h5Index + 3])

    # Do the actual extrapolations
    print("Running extrapolations.")
    stdout.flush()
    # Ws = Waveforms(_vectorW(Ws))
    # Ws.CommonTimeIsSet()
    # print([i for i in range(1)]); stdout.flush()
    # ExtrapolatedWaveformsObject = Ws.extrapolate(Radii, ExtrapolationOrders, Omegas)
    # print(type(ExtrapolatedWaveformsObject))
    # print([10])
    # for i in range(10):
    #     print("Yep"); stdout.flush()
    # print([i for i in range(1)]); stdout.flush()
    # ExtrapolatedWaveforms = [ExtrapolatedWaveformsObject.GetWaveform(i)
    #                         for i in range(ExtrapolatedWaveformsObject.size())]
    ExtrapolatedWaveforms = _Extrapolate(Ws, Radii, ExtrapolationOrders, Omegas, NoiseFloor)

    NExtrapolations = len(ExtrapolationOrders)
    for i, ExtrapolationOrder in enumerate(ExtrapolationOrders):
        # If necessary, rotate
        if OutputFrame == Inertial or OutputFrame == Corotating:
            stdout.write(f"N={ExtrapolationOrder}: Rotating into inertial frame... ")
            stdout.flush()
            ExtrapolatedWaveforms[i].to_inertial_frame()
            print("☺")
            stdout.flush()
        if OutputFrame == Corotating:
            stdout.write(f"N={ExtrapolationOrder}: Rotating into corotating frame... ")
            stdout.flush()
            ExtrapolatedWaveforms[i].to_corotating_frame()
            print("☺")
            stdout.flush()

        # Append the relevant information to the history
        ExtrapolatedWaveforms[i]._append_history(str(InputArguments))
        ExtrapolatedWaveforms[i].extrapolate_coord_radii = CoordRadiiKwarg

        # Output the data
        if OutputDirectory:
            ExtrapolatedFile = OutputDirectory + ExtrapolatedFiles.format(N=ExtrapolationOrder)
            stdout.write(f"N={ExtrapolationOrder}: Writing {ExtrapolatedFile}... ")
            stdout.flush()
            if not exists(OutputDirectory):
                makedirs(OutputDirectory)
            if ExtrapolatedFile.endswith(".dat"):
                ExtrapolatedWaveforms[i].Output(
                    dirname(ExtrapolatedFile)
                    + "/"
                    + ExtrapolatedWaveforms[i].descriptor_string
                    + "_"
                    + ExtrapolatedWaveforms[i].frame_type_string
                    + "_"
                    + basename(ExtrapolatedFile)
                )

            else:
                from scri.SpEC import write_to_h5

                if i == 0:
                    file_write_mode = "w"
                else:
                    file_write_mode = "a"
                write_to_h5(
                    ExtrapolatedWaveforms[i],
                    ExtrapolatedFile,
                    file_write_mode=file_write_mode,
                    use_NRAR_format=UseStupidNRARFormat,
                )
            print("☺")
            stdout.flush()

    if OutputDirectory:
        MaxNormTime = ExtrapolatedWaveforms[0].max_norm_time()
        FileNamePrefixString = (
            ExtrapolatedWaveforms[0].descriptor_string + "_" + ExtrapolatedWaveforms[0].frame_type_string + "_"
        )
        if PlotFormat:
            figabs.gca().set_xlabel(r"$(t-r_\ast)/M$")
            figarg.gca().set_xlabel(r"$(t-r_\ast)/M$")
            fignorm.gca().set_xlabel(r"$(t-r_\ast)/M$")
            figabs.gca().set_ylabel(
                r"$\Delta\, \mathrm{abs} \left( " + ExtrapolatedWaveforms[0].data_type_latex + r" \right) $"
            )
            figarg.gca().set_ylabel(
                r"$\Delta\, \mathrm{uarg} \left( " + ExtrapolatedWaveforms[0].data_type_latex + r" \right) $"
            )
            fignorm.gca().set_ylabel(
                r"$\left\| \Delta\, " + ExtrapolatedWaveforms[0].data_type_latex + r" \right\|_{L_2} $"
            )

        for i, ExtrapolationOrder in reversed(list(enumerate(ExtrapolationOrders))):
            if i > 0:  # Compare to the last one
                if DifferenceFiles or PlotFormat:
                    Diff = scri.WaveformModes(ExtrapolatedWaveforms[i].compare(ExtrapolatedWaveforms[i - 1]))
                if DifferenceFiles:
                    DifferenceFile = OutputDirectory + DifferenceFiles.format(
                        N=ExtrapolationOrder, Nm1=ExtrapolationOrders[i - 1]
                    )
                    stdout.write(f"N={ExtrapolationOrder}: Writing {DifferenceFile}... ")
                    stdout.flush()
                    if DifferenceFile.endswith(".dat"):
                        Diff.Output(
                            dirname(DifferenceFile)
                            + "/"
                            + Diff.descriptor_string
                            + "_"
                            + Diff.frame_type_string
                            + "_"
                            + basename(DifferenceFile)
                        )
                    else:
                        from scri.SpEC import write_to_h5

                        write_to_h5(Diff, DifferenceFile, use_NRAR_format=UseStupidNRARFormat)
                    print("☺")
                    stdout.flush()
                if PlotFormat:
                    # stdout.write("Plotting... "); stdout.flush()
                    Interpolated = scri.WaveformModes(ExtrapolatedWaveforms[i].interpolate(Diff.t))
                    Normalization = Interpolated.norm(True)
                    rep_A = splrep(
                        ExtrapolatedWaveforms[i].t,
                        ExtrapolatedWaveforms[i].abs[:, ExtrapolatedWaveforms[i].index(2, 2)],
                        s=0,
                    )
                    rep_B = splrep(
                        ExtrapolatedWaveforms[i - 1].t,
                        ExtrapolatedWaveforms[i - 1].abs[:, ExtrapolatedWaveforms[i - 1].index(2, 2)],
                        s=0,
                    )
                    AbsA = splev(Diff.t, rep_A, der=0)
                    AbsB = splev(Diff.t, rep_B, der=0)
                    AbsDiff = abs(AbsA - AbsB) / AbsA
                    rep_arg_A = splrep(
                        ExtrapolatedWaveforms[i].t,
                        ExtrapolatedWaveforms[i].arg_unwrapped[:, ExtrapolatedWaveforms[i].index(2, 2)],
                        s=0,
                    )
                    rep_arg_B = splrep(
                        ExtrapolatedWaveforms[i].t,
                        ExtrapolatedWaveforms[i - 1].arg_unwrapped[:, ExtrapolatedWaveforms[i - 1].index(2, 2)],
                        s=0,
                    )
                    ArgDiff = splev(Diff.t, rep_arg_A, der=0) - splev(Diff.t, rep_arg_B, der=0)

                    if abs(ArgDiff[len(ArgDiff) // 3]) > 1.9 * pi:
                        ArgDiff -= 2 * pi * round(ArgDiff[len(ArgDiff) // 3] / (2 * pi))
                    plt.figure(0)
                    plt.semilogy(
                        Diff.t,
                        AbsDiff,
                        label=r"$(N={}) - (N={})$".format(ExtrapolationOrder, ExtrapolationOrders[i - 1]),
                    )
                    plt.figure(1)
                    plt.semilogy(
                        Diff.t,
                        abs(ArgDiff),
                        label=r"$(N={}) - (N={})$".format(ExtrapolationOrder, ExtrapolationOrders[i - 1]),
                    )
                    plt.figure(2)
                    plt.semilogy(
                        Diff.t,
                        Diff.norm(True) / Normalization,
                        label=r"$(N={}) - (N={})$".format(ExtrapolationOrder, ExtrapolationOrders[i - 1]),
                    )
                    # print("☺"); stdout.flush()

        # Finish up the plots and save
        if PlotFormat:
            stdout.write("Saving plots... ")
            stdout.flush()
            plt.figure(0)
            plt.legend(
                borderpad=0.2,
                labelspacing=0.1,
                handlelength=1.5,
                handletextpad=0.1,
                loc="lower left",
                prop={"size": "small"},
            )
            plt.gca().set_ylim(1e-8, 10)
            plt.gca().axvline(x=MaxNormTime, ls="--")
            try:
                from matplotlib.pyplot import tight_layout

                tight_layout(pad=0.5)
            except:
                pass
            figabs.savefig("{}/{}ExtrapConvergence_Abs.{}".format(OutputDirectory, FileNamePrefixString, PlotFormat))
            if PlotFormat != "png":
                figabs.savefig("{}/{}ExtrapConvergence_Abs.{}".format(OutputDirectory, FileNamePrefixString, "png"))
            plt.gca().set_xlim(MaxNormTime - 500.0, MaxNormTime + 200.0)
            figabs.savefig("{}/{}ExtrapConvergence_Abs_Merger.{}".format(OutputDirectory, FileNamePrefixString, PlotFormat))
            if PlotFormat != "png":
                figabs.savefig("{}/{}ExtrapConvergence_Abs_Merger.{}".format(OutputDirectory, FileNamePrefixString, "png"))
            plt.close(figabs)
            plt.figure(1)
            plt.legend(
                borderpad=0.2,
                labelspacing=0.1,
                handlelength=1.5,
                handletextpad=0.1,
                loc="lower left",
                prop={"size": "small"},
            )
            plt.gca().set_xlabel("")
            plt.gca().set_ylim(1e-8, 10)
            plt.gca().axvline(x=MaxNormTime, ls="--")
            try:
                tight_layout(pad=0.5)
            except:
                pass
            figarg.savefig("{}/{}ExtrapConvergence_Arg.{}".format(OutputDirectory, FileNamePrefixString, PlotFormat))
            if PlotFormat != "png":
                figarg.savefig("{}/{}ExtrapConvergence_Arg.{}".format(OutputDirectory, FileNamePrefixString, "png"))
            plt.gca().set_xlim(MaxNormTime - 500.0, MaxNormTime + 200.0)
            figarg.savefig("{}/{}ExtrapConvergence_Arg_Merger.{}".format(OutputDirectory, FileNamePrefixString, PlotFormat))
            if PlotFormat != "png":
                figarg.savefig("{}/{}ExtrapConvergence_Arg_Merger.{}".format(OutputDirectory, FileNamePrefixString, "png"))
            plt.close(figarg)
            plt.figure(2)
            plt.legend(
                borderpad=0.2,
                labelspacing=0.1,
                handlelength=1.5,
                handletextpad=0.1,
                loc="lower left",
                prop={"size": "small"},
            )
            plt.gca().set_ylim(1e-6, 10)
            plt.gca().axvline(x=MaxNormTime, ls="--")
            try:
                tight_layout(pad=0.5)
            except:
                pass
            fignorm.savefig("{}/{}ExtrapConvergence_Norm.{}".format(OutputDirectory, FileNamePrefixString, PlotFormat))
            if PlotFormat != "png":
                fignorm.savefig("{}/{}ExtrapConvergence_Norm.{}".format(OutputDirectory, FileNamePrefixString, "png"))
            plt.gca().set_xlim(MaxNormTime - 500.0, MaxNormTime + 200.0)
            fignorm.savefig(
                "{}/{}ExtrapConvergence_Norm_Merger.{}".format(OutputDirectory, FileNamePrefixString, PlotFormat)
            )
            if PlotFormat != "png":
                fignorm.savefig(
                    "{}/{}ExtrapConvergence_Norm_Merger.{}".format(OutputDirectory, FileNamePrefixString, "png")
                )
            plt.close(fignorm)
            print("☺")
            stdout.flush()

    if return_finite_radius_waveforms:
        return ExtrapolatedWaveforms, Ws
    return ExtrapolatedWaveforms


#####################################
### Batch extrapolation utilities ###
#####################################


# Local utility function
def _safe_format(s, **keys):
    """Like str.format, but doesn't mind missing arguments.

    This function is used to replace strings like '{SomeKey}' in
    the template with the arguments given as keys.  For example,

      _safe_format('{SomeKey} {SomeOtherKey}', SomeKey='Hello', SomeMissingKey='Bla')

    returns 'Hello {SomeOtherKey}', without errors, ignoring the
    `SomeMissingKey` argument, and not bothering with
    '{SomeOtherKey}', so that that can be replaced later.

    """

    class Default(dict):
        def __missing__(self, key):
            return "{" + key + "}"

    from string import Formatter

    return Formatter().vformat(s, (), Default(keys))


def UnstartedExtrapolations(TopLevelOutputDir, SubdirectoriesAndDataFiles):
    """Find unstarted extrapolation directories."""
    from os.path import exists

    Unstarted = []
    for Subdirectory, DataFile in SubdirectoriesAndDataFiles:
        StartedFile = "{}/{}/.started_{}".format(TopLevelOutputDir, Subdirectory, DataFile)
        if not exists(StartedFile):
            Unstarted.append([Subdirectory, DataFile])
    return Unstarted


def NewerDataThanExtrapolation(TopLevelInputDir, TopLevelOutputDir, SubdirectoriesAndDataFiles):
    """Find newer data than extrapolation."""
    from os.path import exists, getmtime

    Newer = []
    for Subdirectory, DataFile in SubdirectoriesAndDataFiles:
        FinishedFile = "{}/{}/.finished_{}".format(TopLevelOutputDir, Subdirectory, DataFile)
        if exists(FinishedFile):
            TimeFinished = getmtime(FinishedFile)
            Timemetadata = getmtime(f"{TopLevelInputDir}/{Subdirectory}/metadata.txt")
            TimeData = getmtime(f"{TopLevelInputDir}/{Subdirectory}/{DataFile}")
            if TimeData > TimeFinished or Timemetadata > TimeFinished:
                Newer.append([Subdirectory, DataFile])
    return Newer


def StartedButUnfinishedExtrapolations(TopLevelOutputDir, SubdirectoriesAndDataFiles):
    """Find directories with extrapolations that started but didn't finish."""
    from os.path import exists

    Unfinished = []
    for Subdirectory, DataFile in SubdirectoriesAndDataFiles:
        StartedFile = "{}/{}/.started_{}".format(TopLevelOutputDir, Subdirectory, DataFile)
        ErrorFile = "{}/{}/.error_{}".format(TopLevelOutputDir, Subdirectory, DataFile)
        FinishedFile = "{}/{}/.finished_{}".format(TopLevelOutputDir, Subdirectory, DataFile)
        if exists(StartedFile) and not exists(ErrorFile) and not exists(FinishedFile):
            Unfinished.append([Subdirectory, DataFile])
    return Unfinished


def ErroredExtrapolations(TopLevelOutputDir, SubdirectoriesAndDataFiles):
    """Find directories with errors."""
    from os.path import exists

    Errored = []
    for Subdirectory, DataFile in SubdirectoriesAndDataFiles:
        ErrorFile = "{}/{}/.error_{}".format(TopLevelOutputDir, Subdirectory, DataFile)
        if exists(ErrorFile):
            Errored.append([Subdirectory, DataFile])
    return Errored


def FindPossibleExtrapolationsToRun(TopLevelInputDir):
    """Find all possible extrapolations."""
    from os import walk
    from re import compile as re_compile

    SubdirectoriesAndDataFiles = []
    LevPattern = re_compile(r"/Lev[0-9]*$")

    # Walk the input directory
    for step in walk(TopLevelInputDir, followlinks=True):
        if LevPattern.search(step[0]):
            if "metadata.txt" in step[2]:
                if "rh_FiniteRadii_CodeUnits.h5" in step[2]:
                    SubdirectoriesAndDataFiles.append(
                        [step[0].replace(TopLevelInputDir + "/", ""), "rh_FiniteRadii_CodeUnits.h5",]
                    )
                if "rPsi4_FiniteRadii_CodeUnits.h5" in step[2]:
                    SubdirectoriesAndDataFiles.append(
                        [step[0].replace(TopLevelInputDir + "/", ""), "rPsi4_FiniteRadii_CodeUnits.h5",]
                    )
                if "r2Psi3_FiniteRadii_CodeUnits.h5" in step[2]:
                    SubdirectoriesAndDataFiles.append(
                        [step[0].replace(TopLevelInputDir + "/", ""), "r2Psi3_FiniteRadii_CodeUnits.h5",]
                    )
                if "r3Psi2_FiniteRadii_CodeUnits.h5" in step[2]:
                    SubdirectoriesAndDataFiles.append(
                        [step[0].replace(TopLevelInputDir + "/", ""), "r3Psi2_FiniteRadii_CodeUnits.h5",]
                    )
                if "r4Psi1_FiniteRadii_CodeUnits.h5" in step[2]:
                    SubdirectoriesAndDataFiles.append(
                        [step[0].replace(TopLevelInputDir + "/", ""), "r4Psi1_FiniteRadii_CodeUnits.h5",]
                    )
                if "r5Psi0_FiniteRadii_CodeUnits.h5" in step[2]:
                    SubdirectoriesAndDataFiles.append(
                        [step[0].replace(TopLevelInputDir + "/", ""), "r5Psi0_FiniteRadii_CodeUnits.h5",]
                    )
    return SubdirectoriesAndDataFiles


def RunExtrapolation(TopLevelInputDir, TopLevelOutputDir, Subdirectory, DataFile, Template):
    from os import makedirs, chdir, getcwd, utime, remove
    from os.path import exists
    from subprocess import call

    InputDir = f"{TopLevelInputDir}/{Subdirectory}"
    OutputDir = f"{TopLevelOutputDir}/{Subdirectory}"
    if not exists(OutputDir):
        makedirs(OutputDir)

    # If OutputDir/.started_r...h5 doesn't exist, touch it; remove errors and
    # finished reports
    with open(f"{OutputDir}/.started_{DataFile}", "a") as f:
        pass
    utime(f"{OutputDir}/.started_{DataFile}", None)
    if exists(f"{OutputDir}/.error_{DataFile}"):
        remove(f"{OutputDir}/.error_{DataFile}")
    if exists(f"{OutputDir}/.finished_{DataFile}"):
        remove(f"{OutputDir}/.finished_{DataFile}")

    # Copy the template file to OutputDir
    with open("{}/Extrapolate_{}.py".format(OutputDir, DataFile[:-3]), "w") as TemplateFile:
        TemplateFile.write(_safe_format(Template, DataFile=DataFile, Subdirectory=Subdirectory))

    # Try to run the extrapolation
    OriginalDir = getcwd()
    try:
        try:
            chdir(OutputDir)
        except:
            print(f"Couldn't change directory to '{OutputDir}'.")
            raise
        print("\n\nRunning {1}/Extrapolate_{0}.py\n\n".format(DataFile[:-3], getcwd()))
        ReturnValue = call(
            "set -o pipefail; python Extrapolate_{0}.py 2>&1 | " + "tee Extrapolate_{}.log".format(DataFile[:-3]),
            shell=True,
        )
        if ReturnValue:
            print(
                "\n\nRunExtrapolation got an error ({4}) on "
                + "['{}', '{}', '{}', '{}'].\n\n".format(
                    TopLevelInputDir,
                    TopLevelOutputDir,
                    Subdirectory,
                    DataFile,
                    ReturnValue,
                )
            )
            with open(f"{OutputDir}/.error_{DataFile}", "w"):
                pass
            chdir(OriginalDir)
            return ReturnValue
        with open(f"{OutputDir}/.finished_{DataFile}", "w"):
            pass
        print("\n\nFinished Extrapolate_{}.py in {}\n\n".format(DataFile[:-3], getcwd()))
    except:
        with open(f"{OutputDir}/.error_{DataFile}", "w"):
            pass
        print(
            "\n\nRunExtrapolation got an error on "
            + "['{}', '{}', '{}', '{}'].\n\n".format(TopLevelInputDir, TopLevelOutputDir, Subdirectory, DataFile)
        )
    finally:
        chdir(OriginalDir)

    return 0


def _Extrapolate(FiniteRadiusWaveforms, Radii, ExtrapolationOrders, Omegas=None, NoiseFloor=None):
    import numpy
    import scri
    from tqdm import trange

    # Get the various dimensions, etc.
    MaxN = max(ExtrapolationOrders)
    MinN = min(ExtrapolationOrders)
    UseOmegas = (Omegas is not None) and (len(Omegas) != 0)
    NTimes = FiniteRadiusWaveforms[0].n_times
    NModes = FiniteRadiusWaveforms[0].n_modes
    NFiniteRadii = len(FiniteRadiusWaveforms)
    NExtrapolations = len(ExtrapolationOrders)
    #SVDTol = 1.0e-12  # Same as Numerical Recipes default in fitsvd.h
    DataType = FiniteRadiusWaveforms[NFiniteRadii - 1].dataType
    ExcludeInsignificantRadii = DataType in [scri.psi1, scri.psi0] and bool(NoiseFloor)
    if ExcludeInsignificantRadii:
        from spherical_functions import LM_index

        ell_min = FiniteRadiusWaveforms[NFiniteRadii - 1].ell_min
        print("Performing extrapolation excluding insignifcant outer radii.")

    # Make sure everyone is playing with a full deck
    if abs(MinN) > NFiniteRadii:
        print(
            """ERROR: Asking for finite-radius waveform {0}, but only got """
            + """{1} finite-radius Waveform objects; need at least as {2} """
            + """finite-radius waveforms.

        """.format(
                MinN, NFiniteRadii, abs(MinN)
            )
        )
        raise ValueError("scri_IndexOutOfBounds")
    if MaxN > 0 and MaxN >= NFiniteRadii:
        print(
            "ERROR: Asking for extrapolation up to order {}, but only got {} "
            "finite-radius Waveform objects; need at least {} waveforms.".format(MaxN, NFiniteRadii, MaxN + 1)
        )
        raise ValueError("scri_IndexOutOfBounds")
    if len(Radii) != NFiniteRadii:
        print(
            """ERROR: Mismatch in data to be extrapolated; there are different """
            + """numbers of waveforms and radius vectors.
        len(FiniteRadiusWaveforms)={}; len(Radii)={}
        """.format(
                NFiniteRadii, len(Radii)
            )
        )
        raise ValueError("scri_VectorSizeMismatch")
    if UseOmegas and len(Omegas) != NTimes:
        print(
            "ERROR: NTimes mismatch in data to be extrapolated.\n"
            + f"       FiniteRadiusWaveforms[0].NTimes()={NTimes}\n"
            + "       Omegas.size()={}\n".format(len(Omegas))
            + "\n"
        )
        raise ValueError("scri_VectorSizeMismatch")
    for i_W in range(1, NFiniteRadii):
        if FiniteRadiusWaveforms[i_W].n_times != NTimes:
            print(
                "ERROR: NTimes mismatch in data to be extrapolated.\n"
                + f"       FiniteRadiusWaveforms[0].NTimes()={NTimes}\n"
                + "       FiniteRadiusWaveforms[{i_W}].NTimes()={0}\n".format(FiniteRadiusWaveforms[i_W].n_times)
                + "\n"
            )
            raise ValueError("scri_VectorSizeMismatch")
        if FiniteRadiusWaveforms[i_W].n_modes != NModes:
            print(
                "ERROR: NModes mismatch in data to be extrapolated.\n"
                + f"       FiniteRadiusWaveforms[0].NModes()={NModes}\n"
                + "       FiniteRadiusWaveforms[{i_W}].NModes()={0}\n".format(FiniteRadiusWaveforms[i_W].n_modes)
                + "\n"
            )
            raise ValueError("scri_VectorSizeMismatch")
        if len(Radii[i_W]) != NTimes:
            print(
                "ERROR: NTimes mismatch in data to be extrapolated.\n"
                + f"       FiniteRadiusWaveforms[0].NTimes()={NTimes}\n"
                + "       Radii[{}].size()={}\n".format(i_W, len(Radii[i_W]))
                + "\n"
            )
            raise ValueError("scri_VectorSizeMismatch")

    # Set up the containers that will be used to store the data during
    # extrapolation.  These are needed so that we don't have to make too many
    # calls to Waveform object methods, which have to go through the _Waveform
    # wrapper, and are therefore slow.
    data = numpy.array([W.data for W in FiniteRadiusWaveforms])
    data = data.view(dtype=float).reshape((data.shape[0], data.shape[1], data.shape[2], 2))
    extrapolated_data = numpy.empty((NExtrapolations, NTimes, NModes), dtype=complex)

    # Set up the output data, recording everything but the mode data
    ExtrapolatedWaveforms = [None] * NExtrapolations
    for i_N in range(NExtrapolations):
        N = ExtrapolationOrders[i_N]
        if N < 0:
            ExtrapolatedWaveforms[i_N] = scri.WaveformModes(FiniteRadiusWaveforms[NFiniteRadii + N])
            ExtrapolatedWaveforms[i_N].history += [f"### Extrapolating with N={N}\n"]
        else:
            # Do everything but set the data
            ExtrapolatedWaveforms[i_N] = scri.WaveformModes(
                t=FiniteRadiusWaveforms[NFiniteRadii - 1].t,
                frame=FiniteRadiusWaveforms[NFiniteRadii - 1].frame,
                data=np.zeros(
                    (NTimes, NModes),
                    dtype=FiniteRadiusWaveforms[NFiniteRadii - 1].data.dtype,
                ),
                history=FiniteRadiusWaveforms[NFiniteRadii - 1].history + [f"### Extrapolating with N={N}\n"],
                version_hist=FiniteRadiusWaveforms[NFiniteRadii - 1].version_hist,
                frameType=FiniteRadiusWaveforms[NFiniteRadii - 1].frameType,
                dataType=FiniteRadiusWaveforms[NFiniteRadii - 1].dataType,
                r_is_scaled_out=FiniteRadiusWaveforms[NFiniteRadii - 1].r_is_scaled_out,
                m_is_scaled_out=FiniteRadiusWaveforms[NFiniteRadii - 1].m_is_scaled_out,
                ell_min=FiniteRadiusWaveforms[NFiniteRadii - 1].ell_min,
                ell_max=FiniteRadiusWaveforms[NFiniteRadii - 1].ell_max,
            )

    if MaxN < 0:
        return ExtrapolatedWaveforms

    MaxCoefficients = MaxN + 1

    # Loop over time
    for i_t in trange(NTimes):

        # Set up the radius data (if we are NOT using Omega)
        if not UseOmegas:
            OneOverRadii = [1.0 / Radii[i_W][i_t] for i_W in range(NFiniteRadii)]

            data_t = data[:, i_t, :, 0] + 1j * data[:, i_t, :, 1]

            if ExcludeInsignificantRadii:
                # For the sake of avoiding a poorly conditioned polynomial fit, we need to set an absolute minimum for the
                # possible number of radii that may be used. Occasionally where there is junk radiation the amplitude of
                # the waveform may be spuriously small at a few timesteps, setting this minimum avoids excluding an
                # unreasonable number of radii and then having polyfit return an error.
                MinimumNRadii = max(ExtrapolationOrders) + 4

                # Since we are in the corotating frame, we can be sure that the (2,2) mode is dominant. For the sake of
                # consistency we will use the same radial weights for each mode at a given retarded time.
                WaveformAmplitude = np.linalg.norm(Re[MinimumNRadii] + 1j * Im[MinimumNRadii])

                # Radius at which the amplitude of the waveform is equal to error_tol.
                LargestSignificantRadius = (WaveformAmplitude / NoiseFloor) ** (1 / (6 - DataType))

                # Use as many radii as possible that are smaller than LargestSignificantRadius.
                RadiiOnTimeSlice = np.array(Radii)[:, i_t]
                MinimumNRadii = max(sum(RadiiOnTimeSlice <= LargestSignificantRadius), MinimumNRadii)

                # Remove the outer radii
                OneOverRadii = OneOverRadii[:MinimumNRadii]
                data_t = data_t[:MinimumNRadii, :]

            # Loop over extrapolation orders
            for i_N in range(NExtrapolations):
                N = ExtrapolationOrders[i_N]

                # If non-extrapolating, skip to the next one (the copying was
                # done when ExtrapolatedWaveforms[i_N] was constructed)
                if N < 0:
                    continue

                # Do the extrapolations
                extrapolated_data[i_N, i_t, :] = polyfit(OneOverRadii, data_t, N)[0, :]

        else:  # UseOmegas

            # Loop over modes
            for i_m in range(NModes):

                # Set up the radius data (if we ARE using Omega)
                M = FiniteRadiusWaveforms[0].LM(i_m)[1]
                if UseOmegas:
                    OneOverRadii = [
                        1.0 / (Radii[i_W][i_t] * M * Omegas[i_t]) if M != 0 else 1.0 / (Radii[i_W][i_t])
                        for i_W in range(NFiniteRadii)
                    ]

                # Set up the mode data
                Re = data[:, i_m, i_t, 0]
                Im = data[:, i_m, i_t, 1]

                # Loop over extrapolation orders
                for i_N in range(NExtrapolations):
                    N = ExtrapolationOrders[i_N]

                    # If non-extrapolating, skip to the next one (the copying was
                    # done when ExtrapolatedWaveforms[i_N] was constructed)
                    if N < 0:
                        continue

                    # Do the extrapolations
                    re = numpy.polyfit(OneOverRadii, Re, N)[-1]
                    im = numpy.polyfit(OneOverRadii, Im, N)[-1]
                    extrapolated_data[i_N, i_t, i_m] = re + 1j * im

    for i_N in range(NExtrapolations):
        N = ExtrapolationOrders[i_N]
        if N >= 0:
            ExtrapolatedWaveforms[i_N].data[...] = extrapolated_data[i_N]

    print("")

    return ExtrapolatedWaveforms



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import pytest
import os.path
import numpy as np
import h5py
import scri
import scri.SpEC

# NOTE: if test_file_io() comes after test_NRAR_extrapolation() in this file, then the
# output of the latter can be used in test_file_io(). Otherwise, test_file_io() will
# just generate the data it needs indepenedently.


@pytest.fixture(scope="session")  # The following will exist for an entire run of pytest
def tempdir(tmp_path_factory):
    """Test fixture to create a temp directory and write the fake finite-radius file"""
    tmpdir = tmp_path_factory.mktemp("test_extrapolation")
    filename = tmpdir / "rh_FiniteRadii_CodeUnits.h5"
    scri.sample_waveforms.create_fake_finite_radius_strain_h5file(
        output_file_path=filename,
        ell_max=3,
        t_1=2000.0,
    )
    return tmpdir


def test_extrapolation(tempdir):
    """Test extrapolation of waveforms with output in the scri format"""
    input_dir = str(tempdir)
    output_dir = str(tempdir / "test_extrapolation")
    scri.extrapolation.extrapolate(InputDirectory=input_dir, OutputDirectory=output_dir, ChMass=1.0)


def test_NRAR_extrapolation(tempdir):
    """Test extrapolation of waveforms with output in the NRAR format."""
    input_dir = str(tempdir)
    output_dir = str(tempdir / "test_NRAR_extrapolation")
    scri.extrapolation.extrapolate(
        InputDirectory=input_dir,
        OutputDirectory=output_dir,
        ChMass=1.0,
        UseStupidNRARFormat=True,
        DifferenceFiles="",  # Computing DifferenceFiles and making plots are tested in test_extrapolation() so
        PlotFormat="",
    )  # to save time we don't do it here.


def test_file_io(tempdir):
    """Test file I/O with a round-trip and compare H5 files"""
    output_dir = tempdir / "test_file_io"
    input_file = tempdir / "test_NRAR_extrapolation" / "rhOverM_Extrapolated_N2.h5"
    output_file = str(output_dir / "Asymptotic_GeometricUnits.h5")
    output_file_result = str(output_dir / "rhOverM_Asymptotic_GeometricUnits.h5")

    if not input_file.exists():
        scri.extrapolation.extrapolate(
            InputDirectory=str(tempdir),
            OutputDirectory=str(tempdir / "test_NRAR_extrapolation"),
            ChMass=1.0,
            UseStupidNRARFormat=True,
            DifferenceFiles="",
            PlotFormat="",
        )
    output_dir.mkdir()

    input_file = str(input_file)
    w = scri.SpEC.read_from_h5(input_file)
    scri.SpEC.write_to_h5(w, output_file)

    with h5py.File(input_file, "r") as f1, h5py.File(output_file_result, "r") as f2:
        # Check top-level attributes
        list1 = sorted(f1.attrs.items())
        list2 = sorted(f2.attrs.items())
        # attribute 3 is waveform output version
        list1.remove(list1[3])
        list2.remove(list2[3])
        assert list1 == list2

        # Check that original history is contained (with extra comment characters) in second history
        if isinstance(f2["History.txt"][()], str):
            assert ("# " + "\n# ".join(f1["History.txt"][()].split("\n"))) in f2["History.txt"][()]
        else:
            assert ("# " + "\n# ".join(f1["History.txt"][()].decode("utf-8").split("\n"))) in f2["History.txt"][()].decode("utf-8")

        # Now check each mode from the original
        for mode in list(f1):
            if mode.startswith("Y"):
                assert list(f1[mode].attrs.items()) == list(f2[mode].attrs.items())
                assert np.array_equal(f1[mode][:], f2[mode][:])



import copy
import h5py
import scipy
import spinsfast
from scri.asymptotic_bondi_data.transformations import _process_transformation_kwargs, boosted_grid, conformal_factors
import numpy as np
import spherical_functions as sf


def fourvec_to_spin_matrix(fourvec):
    """Inner product of a four vector and the Pauli matrices,
    as defined by Eq. (1.2.39) of Spinors and Spacetime Vol. 1

    Parameters
    ----------
    fourvec: float array of length 4.
    """

    a = np.cos(fourvec[0] / 2) + 1j * fourvec[3] * np.sin(fourvec[0] / 2)
    b = (-fourvec[2] + 1j * fourvec[1]) * np.sin(fourvec[0] / 2)
    c = (+fourvec[2] + 1j * fourvec[1]) * np.sin(fourvec[0] / 2)
    d = np.cos(fourvec[0] / 2) - 1j * fourvec[3] * np.sin(fourvec[0] / 2)

    return np.array([[a, b], [c, d]])


def Lorentz_to_spin_matrix(lorentz):
    """Convert a Lorentz transformation to a spin matrix.

    Parameters
    ----------
    lorentz: LorentzTransformation
    """

    psi = 2 * np.arctan2(np.linalg.norm(lorentz.frame_rotation.components[1:]), lorentz.frame_rotation.components[0])
    if psi == 0:
        frame_rotation_vec_hat = np.array([0, 0, 0])
    else:
        frame_rotation_vec_hat = lorentz.frame_rotation.components[1:] / np.linalg.norm(
            lorentz.frame_rotation.components[1:]
        )

    chi = 1j * np.arctanh(np.linalg.norm(lorentz.boost_velocity))
    if chi == 0:
        boost_velocity_vec_hat = np.array([0, 0, 0])
    else:
        boost_velocity_vec_hat = lorentz.boost_velocity / np.linalg.norm(lorentz.boost_velocity)

    # compute product of Lorentz vector and Pauli matrices
    frame_rotation_spin_matrix = fourvec_to_spin_matrix([psi, *frame_rotation_vec_hat])
    boost_velocity_spin_matrix = fourvec_to_spin_matrix([chi, *boost_velocity_vec_hat])

    if lorentz.order.index("frame_rotation") < lorentz.order.index("boost_velocity"):
        return np.matmul(boost_velocity_spin_matrix, frame_rotation_spin_matrix)
    else:
        return np.matmul(frame_rotation_spin_matrix, boost_velocity_spin_matrix)


def pure_spin_matrix_to_Lorentz(A, is_rotation=None, tol=1e-14):
    """Convert a pure spin matrix to a rotation or a boost.

    Parameters
    ----------
    A: 2 by 2 complex array
        2 by 2 array corresponding to the input spin matrix.
    is_rotation: bool
        Whether or not the spin matrix should correspond to
        a Lorentz rotation or a Lorentz boost. Defaults to None,
        i.e., this will be figured out based on the matrix's values.
    tol: float
        Tolerance for figuring out whether the spin matrix
        corresponds to a Lorentz rotation or a Lorentz boost.
    """
    logA = scipy.linalg.logm(A)

    # compute the Lorentz vector
    nvec = np.array([(logA[1, 0] + logA[0, 1]) / 2, (logA[1, 0] - logA[0, 1]) / 2, (logA[0, 0] - logA[1, 1]) / 2])

    nvec_re = np.array([nvec[0].imag, nvec[1].real, nvec[2].imag])
    nvec_im = np.array([-nvec[0].real, nvec[1].imag, -nvec[2].real])

    # figure out if this spin matrix is a rotation or a boost
    if is_rotation is None:
        if np.linalg.norm(nvec_im) < tol:
            nvec = nvec_re
            is_rotation = True
        elif np.linalg.norm(nvec_re) < tol:
            nvec = nvec_im
            is_rotation = False
        else:
            spin_matrix_to_Lorentz(A)
    else:
        if is_rotation:
            nvec = nvec_re
        else:
            nvec = nvec_im

    psi = 2 * np.linalg.norm(nvec)
    if psi == 0:
        nvec_hat = np.array([0, 0, 1])
    else:
        nvec_hat = nvec / np.linalg.norm(nvec)

    if is_rotation:
        return np.quaternion(np.cos(psi / 2), *nvec_hat * np.sin(psi / 2))
    else:
        return np.tanh(psi) * nvec_hat


def spin_matrix_to_Lorentz(A, output_order=["frame_rotation", "boost_velocity"]):
    """Convert a spin matrix to a Lorentz transformation.

    This uses SVD to decompose the spin matrix into a
    spin matrix describing a Lorentz boost (positive or negative definite Hermitian) and a
    spin matrix describing a Lorentz rotation (unitary).

    Parameters
    ----------
    A: 2 by 2 complex array
        2 by 2 array corresponding to the input spin matrix.
    output_order: list
        Order in which rotation and boost should be applied.
    """
    if np.allclose(A, np.zeros_like(A)):
        return LorentzTransformation()

    u, s, vh = np.linalg.svd(A)

    try:
        frame_rotation_idx = output_order.index("frame_rotation")
    except:
        frame_rotation_idx = np.inf
    try:
        boost_velocity_idx = output_order.index("boost_velocity")
    except:
        boost_velocity_idx = np.inf

    if frame_rotation_idx < boost_velocity_idx:
        frame_rotation_matrix = np.matmul(u, vh)
        boost_velocity_matrix = np.linalg.multi_dot([u, np.diag(s), u.conj().T])
    else:
        frame_rotation_matrix = np.matmul(u, vh)
        boost_velocity_matrix = np.linalg.multi_dot([vh.conj().T, np.diag(s), vh])
    frame_rotation = pure_spin_matrix_to_Lorentz(frame_rotation_matrix, is_rotation=True)
    boost_velocity = pure_spin_matrix_to_Lorentz(boost_velocity_matrix, is_rotation=False)
    return LorentzTransformation(
        frame_rotation=frame_rotation.components, boost_velocity=boost_velocity, order=output_order
    )


def transform_supertranslation(S, lorentz, ell_max=None):
    """Apply a Lorentz transformation to a supertranslation and multiply by
    the conformal factor. This produces the supertranslation the appears
    when commuting a supertranslation through a Lorentz transformation.

    The Lorentz transformation is the transformation appearing on
    the RHS of the product, i.e., S' = L^{-1} S L.

    Parameters
    ----------
    S: ndarray, dtype=complex
        supertranslation to be transformed.
    lorentz: LorentzTransformation
        Lorentz transformation to be used to transform the supertranslation.
    ell_max: int
        Maximum ell to use when expressing functions via coordinates on the two-sphere.
    """
    if ell_max is None:
        ell_max = lorentz.ell_max
    n_theta = 2 * ell_max + 1
    n_phi = n_theta

    lorentz_inv = lorentz.inverse(output_order=["frame_rotation", "boost_velocity"])

    distorted_grid_rotors = boosted_grid(lorentz_inv.frame_rotation, lorentz_inv.boost_velocity, n_theta, n_phi)
    k, ðk_over_k, one_over_k, one_over_k_cubed = conformal_factors(lorentz_inv.boost_velocity, distorted_grid_rotors)

    return spinsfast.map2salm(
        (k[0] * sf.Grid(sf.Modes(S, spin_weight=0).evaluate(distorted_grid_rotors), spin_weight=0)).real, 0, ell_max
    )


class LorentzTransformation:
    def __init__(self, **kwargs):
        self.ell_max = copy.deepcopy(kwargs.pop("ell_max", 12))
        (
            frame_rotation,
            boost_velocity,
            supertranslation,
            working_ell_max,
            output_ell_max,
        ) = _process_transformation_kwargs(self.ell_max, **kwargs)
        self.frame_rotation = copy.deepcopy(frame_rotation)
        self.boost_velocity = copy.deepcopy(boost_velocity)

        self.order = copy.deepcopy(kwargs.pop("order", ["frame_rotation", "boost_velocity"]))
        if "supertranslation" in self.order:
            del self.order[self.order.index("supertranslation")]
        for transformation in ["frame_rotation", "boost_velocity"]:
            if transformation not in self.order:
                self.order.append(transformation)

    def __repr__(self):
        Lorentz_output = {}
        for transformation in self.order:
            if transformation == "frame_rotation":
                Lorentz_output[transformation] = self.frame_rotation
            elif transformation == "boost_velocity":
                Lorentz_output[transformation] = self.boost_velocity

        return f"LorentzTransformation(\n\t{self.order[0]}={Lorentz_output[self.order[0]]}\n\t{self.order[1]}={Lorentz_output[self.order[1]]}\n)"

    def copy(self):
        return LorentzTransformation(
            frame_rotation=self.frame_rotation.components,
            boost_velocity=self.boost_velocity,
            order=self.order,
            ell_max=self.ell_max,
        )

    def reorder(self, output_order):
        """Reorder a Lorentz transformation."""
        if not ("frame_rotation" in output_order and "boost_velocity" in output_order):
            raise ValueError("Not enough transformations")

        if self.order == output_order:
            return self.copy()
        else:
            L_spin_matrix = Lorentz_to_spin_matrix(self)
            L_reordered = spin_matrix_to_Lorentz(L_spin_matrix, output_order=output_order)
            return L_reordered

    def inverse(self, output_order=None):
        """Compute the inverse of a Lorentz transformation."""
        if output_order is None:
            output_order = self.order[::-1]

        L_spin_matrix = Lorentz_to_spin_matrix(self)
        L_spin_matrix_inverse = np.linalg.inv(L_spin_matrix)
        L_inverse = spin_matrix_to_Lorentz(L_spin_matrix_inverse, output_order=output_order)
        return L_inverse

    def is_close_to(self, other):
        frame_rotation_eq = np.allclose(self.frame_rotation.components, other.frame_rotation.components)
        boost_velocity_eq = np.allclose(self.boost_velocity, other.boost_velocity)
        return frame_rotation_eq and boost_velocity_eq

    def __mul__(self, other):
        """Compose two Lorentz transformations.

        These are composed as other * self, when each is viewed as a (passive)
        Lorentz transformation on the coordinates.

        The output order of the individual transformations is "frame_rotation", "boost_velocity".

        Parameters
        ----------
        other: LorentzTransformation
            2nd Lorentz transformation to be applied.
        """
        L1_spin_matrix = Lorentz_to_spin_matrix(self)
        L2_spin_matrix = Lorentz_to_spin_matrix(other)

        Ls_spin_matrix = np.matmul(L2_spin_matrix, L1_spin_matrix)

        return spin_matrix_to_Lorentz(Ls_spin_matrix, output_order=["frame_rotation", "boost_velocity"])


class BMSTransformation:
    def __init__(self, **kwargs):
        self.ell_max = copy.deepcopy(kwargs.pop("ell_max", 12))
        (
            frame_rotation,
            boost_velocity,
            supertranslation,
            working_ell_max,
            output_ell_max,
        ) = _process_transformation_kwargs(self.ell_max, **kwargs)
        self.frame_rotation = copy.deepcopy(frame_rotation)
        self.boost_velocity = copy.deepcopy(boost_velocity)
        self.supertranslation = np.pad(supertranslation, (0, (self.ell_max + 1) ** 2 - supertranslation.size))

        self.order = copy.deepcopy(kwargs.pop("order", ["supertranslation", "frame_rotation", "boost_velocity"]))
        for transformation in ["supertranslation", "frame_rotation", "boost_velocity"]:
            if transformation not in self.order:
                self.order.append(transformation)

    def __repr__(self):
        BMS_output = {}
        for transformation in self.order:
            if transformation == "frame_rotation":
                BMS_output[transformation] = self.frame_rotation
            elif transformation == "boost_velocity":
                BMS_output[transformation] = self.boost_velocity
            elif transformation == "supertranslation":
                BMS_output[transformation] = self.supertranslation[:9]

        return f"BMSTransformation(\n\t{self.order[0]}={BMS_output[self.order[0]]}\n\t{self.order[1]}={BMS_output[self.order[1]]}\n\t{self.order[2]}={BMS_output[self.order[2]]}\n)"

    def copy(self):
        return BMSTransformation(
            frame_rotation=self.frame_rotation.components,
            boost_velocity=self.boost_velocity,
            supertranslation=self.supertranslation,
            order=self.order,
            ell_max=self.ell_max,
        )

    def reorder(self, output_order):
        """Reorder a BMS transformation."""
        if not (
            "supertranslation" in output_order and "frame_rotation" in output_order and "boost_velocity" in output_order
        ):
            raise ValueError("Not enough transformations")

        # map to normal order
        normal_order = ["supertranslation", "frame_rotation", "boost_velocity"]
        if self.order == normal_order:
            BMS_normal_order = self.copy()
        elif self.order == ["frame_rotation", "supertranslation", "boost_velocity"]:
            S_prime = transform_supertranslation(
                self.supertranslation,
                LorentzTransformation(
                    frame_rotation=self.frame_rotation.components, ell_max=self.ell_max, order=self.order
                ),
            )

            BMS_normal_order = BMSTransformation(
                frame_rotation=self.frame_rotation.components,
                boost_velocity=self.boost_velocity,
                supertranslation=S_prime,
                ell_max=self.ell_max,
                order=normal_order,
            )
        elif self.order == ["frame_rotation", "boost_velocity", "supertranslation"]:
            S_prime = transform_supertranslation(
                self.supertranslation,
                LorentzTransformation(
                    frame_rotation=self.frame_rotation.components,
                    boost_velocity=self.boost_velocity,
                    ell_max=self.ell_max,
                    order=self.order,
                ),
            )

            BMS_normal_order = BMSTransformation(
                frame_rotation=self.frame_rotation.components,
                boost_velocity=self.boost_velocity,
                supertranslation=S_prime,
                ell_max=self.ell_max,
                order=normal_order,
            )
        elif self.order == ["supertranslation", "boost_velocity", "frame_rotation"]:
            L_prime = LorentzTransformation(
                frame_rotation=self.frame_rotation.components,
                boost_velocity=self.boost_velocity,
                ell_max=self.ell_max,
                order=self.order,
            ).reorder(normal_order)

            BMS_normal_order = BMSTransformation(
                frame_rotation=L_prime.frame_rotation.components,
                boost_velocity=L_prime.boost_velocity,
                supertranslation=self.supertranslation,
                ell_max=self.ell_max,
                order=normal_order,
            )
        elif self.order == ["boost_velocity", "supertranslation", "frame_rotation"]:
            L_prime = LorentzTransformation(
                frame_rotation=self.frame_rotation.components,
                boost_velocity=self.boost_velocity,
                ell_max=self.ell_max,
                order=self.order,
            ).reorder(normal_order)

            S_prime = transform_supertranslation(
                self.supertranslation,
                LorentzTransformation(boost_velocity=self.boost_velocity, ell_max=self.ell_max, order=self.order),
            )

            BMS_normal_order = BMSTransformation(
                frame_rotation=L_prime.frame_rotation.components,
                boost_velocity=L_prime.boost_velocity,
                supertranslation=S_prime,
                ell_max=self.ell_max,
                order=normal_order,
            )
        elif self.order == ["boost_velocity", "frame_rotation", "supertranslation"]:
            L_prime = LorentzTransformation(
                frame_rotation=self.frame_rotation.components,
                boost_velocity=self.boost_velocity,
                ell_max=self.ell_max,
                order=self.order,
            ).reorder(normal_order)

            S_prime = transform_supertranslation(self.supertranslation, L_prime)

            BMS_normal_order = BMSTransformation(
                frame_rotation=L_prime.frame_rotation.components,
                boost_velocity=L_prime.boost_velocity,
                supertranslation=S_prime,
                ell_max=self.ell_max,
                order=normal_order,
            )

        # map to output order
        if output_order == normal_order:
            BMS_reordered = BMS_normal_order.copy()
        elif output_order == ["frame_rotation", "supertranslation", "boost_velocity"]:
            S_prime = transform_supertranslation(
                BMS_normal_order.supertranslation,
                LorentzTransformation(
                    frame_rotation=BMS_normal_order.frame_rotation.components,
                    ell_max=BMS_normal_order.ell_max,
                    order=BMS_normal_order.order,
                ).inverse(),
            )

            BMS_reordered = BMSTransformation(
                frame_rotation=BMS_normal_order.frame_rotation.components,
                boost_velocity=BMS_normal_order.boost_velocity,
                supertranslation=S_prime,
                ell_max=BMS_normal_order.ell_max,
                order=output_order,
            )
        elif output_order == ["frame_rotation", "boost_velocity", "supertranslation"]:
            S_prime = transform_supertranslation(
                BMS_normal_order.supertranslation,
                LorentzTransformation(
                    frame_rotation=BMS_normal_order.frame_rotation.components,
                    boost_velocity=BMS_normal_order.boost_velocity,
                    ell_max=BMS_normal_order.ell_max,
                    order=BMS_normal_order.order,
                ).inverse(),
            )

            BMS_reordered = BMSTransformation(
                frame_rotation=BMS_normal_order.frame_rotation.components,
                boost_velocity=BMS_normal_order.boost_velocity,
                supertranslation=S_prime,
                ell_max=BMS_normal_order.ell_max,
                order=output_order,
            )
        elif output_order == ["supertranslation", "boost_velocity", "frame_rotation"]:
            L_prime = LorentzTransformation(
                frame_rotation=BMS_normal_order.frame_rotation.components,
                boost_velocity=BMS_normal_order.boost_velocity,
                ell_max=BMS_normal_order.ell_max,
                order=BMS_normal_order.order,
            ).reorder(output_order)

            BMS_reordered = BMSTransformation(
                frame_rotation=L_prime.frame_rotation.components,
                boost_velocity=L_prime.boost_velocity,
                supertranslation=BMS_normal_order.supertranslation,
                ell_max=BMS_normal_order.ell_max,
                order=output_order,
            )
        elif output_order == ["boost_velocity", "supertranslation", "frame_rotation"]:
            L_prime = LorentzTransformation(
                frame_rotation=BMS_normal_order.frame_rotation.components,
                boost_velocity=BMS_normal_order.boost_velocity,
                ell_max=BMS_normal_order.ell_max,
                order=BMS_normal_order.order,
            ).reorder(output_order)

            S_prime = transform_supertranslation(
                BMS_normal_order.supertranslation,
                LorentzTransformation(
                    boost_velocity=L_prime.boost_velocity, ell_max=L_prime.ell_max, order=L_prime.order
                ).inverse(),
            )

            BMS_reordered = BMSTransformation(
                frame_rotation=L_prime.frame_rotation.components,
                boost_velocity=L_prime.boost_velocity,
                supertranslation=S_prime,
                ell_max=BMS_normal_order.ell_max,
                order=output_order,
            )
        elif output_order == ["boost_velocity", "frame_rotation", "supertranslation"]:
            L_prime = LorentzTransformation(
                frame_rotation=BMS_normal_order.frame_rotation.components,
                boost_velocity=BMS_normal_order.boost_velocity,
                ell_max=BMS_normal_order.ell_max,
                order=BMS_normal_order.order,
            ).reorder(output_order)

            S_prime = transform_supertranslation(BMS_normal_order.supertranslation, L_prime.inverse())

            BMS_reordered = BMSTransformation(
                frame_rotation=L_prime.frame_rotation.components,
                boost_velocity=L_prime.boost_velocity,
                supertranslation=S_prime,
                ell_max=BMS_normal_order.ell_max,
                order=output_order,
            )

        return BMS_reordered

    def inverse(self, output_order=None):
        """Compute the inverse of a BMS transformation."""
        if output_order is None:
            output_order = self.order[::-1]

        bms_normal_order = self.reorder(output_order=["supertranslation", "frame_rotation", "boost_velocity"])

        L_inverse = LorentzTransformation(
            frame_rotation=bms_normal_order.frame_rotation.components,
            boost_velocity=bms_normal_order.boost_velocity,
            ell_max=bms_normal_order.ell_max,
            output_order=bms_normal_order.order,
        ).inverse(output_order=["frame_rotation", "boost_velocity"][::-1])

        n_theta = 2 * self.ell_max + 1
        n_phi = n_theta

        S_inverse = -bms_normal_order.supertranslation

        bms_inverse = BMSTransformation(
            frame_rotation=L_inverse.frame_rotation.components,
            boost_velocity=L_inverse.boost_velocity,
            supertranslation=S_inverse,
            ell_max=bms_normal_order.ell_max,
            order=["supertranslation", "frame_rotation", "boost_velocity"][::-1],
        )

        if bms_inverse.order == output_order:
            return bms_inverse
        else:
            return bms_inverse.reorder(output_order=output_order)

    def is_close_to(self, other):
        frame_rotation_eq = np.allclose(self.frame_rotation.components, other.frame_rotation.components)
        boost_velocity_eq = np.allclose(self.boost_velocity, other.boost_velocity)
        supertranslation_eq = np.allclose(self.supertranslation, other.supertranslation)
        return frame_rotation_eq and boost_velocity_eq and supertranslation_eq

    def __mul__(self, other):
        """Compose two BMS transformations.

        These are composed as other * self, when each is viewed as a (passive)
        BMS transformation on the coordinates.

        The output order of the individual transformations is "supertranslation", "frame_rotation", "boost_velocity".

        For more on this, see the documentation.

        Parameters
        ----------
        other: BMSTransformation
            2nd BMS transformation to be applied.
        """
        ell_max = max(self.ell_max, other.ell_max)

        bms1_normal_order = self.reorder(output_order=["supertranslation", "frame_rotation", "boost_velocity"])
        bms2_normal_order = other.reorder(output_order=["supertranslation", "frame_rotation", "boost_velocity"])

        L1 = LorentzTransformation(
            frame_rotation=bms1_normal_order.frame_rotation.components,
            boost_velocity=bms1_normal_order.boost_velocity,
            ell_max=ell_max,
            order=bms1_normal_order.order,
        )
        L2 = LorentzTransformation(
            frame_rotation=bms2_normal_order.frame_rotation.components,
            boost_velocity=bms2_normal_order.boost_velocity,
            ell_max=ell_max,
            order=bms2_normal_order.order,
        )
        L_composed = L1 * L2

        S1 = np.pad(
            bms1_normal_order.supertranslation, (0, (ell_max + 1) ** 2 - bms1_normal_order.supertranslation.shape[0])
        )
        S2 = np.pad(
            bms2_normal_order.supertranslation, (0, (ell_max + 1) ** 2 - bms2_normal_order.supertranslation.shape[0])
        )

        S_prime = transform_supertranslation(S1, L2)

        S_composed = S_prime + S2

        bms_composed = BMSTransformation(
            frame_rotation=L_composed.frame_rotation.components,
            boost_velocity=L_composed.boost_velocity,
            supertranslation=S_composed,
            ell_max=ell_max,
            order=["supertranslation", "frame_rotation", "boost_velocity"],
        )

        return bms_composed

    def to_file(self, filename, file_write_mode="w", group=None):
        dt = h5py.special_dtype(vlen=str)
        with h5py.File(filename, file_write_mode) as hf:
            if group is not None:
                g = hf.create_group(group)
            else:
                g = hf
            g.create_dataset("supertranslation", data=self.supertranslation)
            g.create_dataset("frame_rotation", data=self.frame_rotation.components)
            g.create_dataset("boost_velocity", data=self.boost_velocity)
            g.create_dataset("order", data=self.order)
            g.create_dataset("ell_max", data=self.ell_max)

        return

    def from_file(self, filename, group=None):
        with h5py.File(filename, "r") as hf:
            if group is not None:
                g = hf[group]
            else:
                g = hf
            supertranslation = np.array(g.get("supertranslation"))
            frame_rotation = np.array(g.get("frame_rotation"))
            boost_velocity = np.array(g.get("boost_velocity"))
            order = [x.decode("utf-8") for x in np.array(g.get("order"))]
            ell_max = int(np.array(g.get("ell_max")))

        BMS = BMSTransformation(
            frame_rotation=frame_rotation,
            boost_velocity=boost_velocity,
            supertranslation=supertranslation,
            order=order,
            ell_max=ell_max,
        )

        self.__dict__.update(BMS.__dict__)

        return



import scri
import pytest
import numpy as np
import spinsfast
import spherical_functions as sf
import quaternion
from quaternion.calculus import derivative

ABD = scri.AsymptoticBondiData


def kerr_schild(mass, spin, ell_max=8):
    psi2 = np.zeros(sf.LM_total_size(0, ell_max), dtype=complex)
    psi1 = np.zeros(sf.LM_total_size(0, ell_max), dtype=complex)
    psi0 = np.zeros(sf.LM_total_size(0, ell_max), dtype=complex)

    # In the Moreschi-Boyle convention
    psi2[0] = -sf.constant_as_ell_0_mode(mass)
    psi1[2] = -np.sqrt(2) * (3j * spin / 2) * (np.sqrt((8 / 3) * np.pi))
    psi0[6] = 2 * (3 * spin**2 / mass / 2) * (np.sqrt((32 / 15) * np.pi))

    return psi2, psi1, psi0


def test_abd_kerr_superrest_frame():
    mass = 2.0
    spin = 0.456
    ell_max = 8
    u = np.linspace(-100, 100, num=5000)
    psi2, psi1, psi0 = kerr_schild(mass, spin, ell_max)
    abd = ABD.from_initial_values(u, ell_max=ell_max, psi2=psi2, psi1=psi1)

    tolerance = 1e-12

    supertranslation = np.array(
        [0.0, 3e-2 - 1j * 5e-3, 1e-3, -3e-2 - 1j * 5e-3, 2e-4 + 1j * 1e-4, 1j * 3e-3, 1e-2, 1j * 3e-3, 2e-4 - 1j * 1e-4]
    )
    frame_rotation = quaternion.quaternion(1, 2, 3, 4).normalized().components
    boost_velocity = np.array([2e-4, -3e-5, 2e-4])
    abd_prime = abd.transform(
        supertranslation=supertranslation, frame_rotation=frame_rotation, boost_velocity=boost_velocity
    )

    abd_recovered, transformation, rel_errs = abd_prime.map_to_superrest_frame(
        t_0=0,
        padding_time=20,
        N_itr_maxes={"superrest": 1, "CoM_transformation": 10, "rotation": 10, "supertranslation": 10},
    )

    PsiM_recovered = abd_recovered.supermomentum("Moreschi")[np.argmin(abs(abd_recovered.t))]
    PsiM_recovered[0:4] = 0
    PsiM_recovered_S2 = spinsfast.salm2map(
        PsiM_recovered.view(np.ndarray), 0, ell_max, 2 * ell_max + 1, 2 * ell_max + 1
    )
    PsiM_recovered_norm = spinsfast.map2salm((abs(PsiM_recovered_S2) ** 2).view(np.ndarray), 0, ell_max)[0] / np.sqrt(
        4 * np.pi
    )
    assert np.allclose(PsiM_recovered_norm, 0.0, atol=tolerance, rtol=tolerance)

    chi_recovered = abd_recovered.bondi_dimensionless_spin()
    chi_recovered = chi_recovered / np.linalg.norm(chi_recovered, axis=-1)[:, None]
    assert np.allclose(chi_recovered, [[0, 0, 1]] * abd_recovered.t.size, atol=tolerance, rtol=tolerance)

    G_recovered = abd_recovered.bondi_CoM_charge() / abd_recovered.bondi_four_momentum()[:, 0, None]
    assert np.allclose(G_recovered[np.argmin(abs(abd_recovered.t))], 0.0, atol=tolerance, rtol=tolerance)
    assert np.allclose(
        derivative(G_recovered, abd_recovered.t)[np.argmin(abs(abd_recovered.t))], 0.0, atol=tolerance, rtol=tolerance
    )


def test_abd_kerr_target_superrest_frame():
    mass = 2.0
    spin = 0.456
    ell_max = 8
    u = np.linspace(-100, 100, num=5000)
    psi2, psi1, psi0 = kerr_schild(mass, spin, ell_max)
    abd = ABD.from_initial_values(u, ell_max=ell_max, psi2=psi2, psi1=psi1)

    tolerance = 1e-12

    supertranslation = np.array(
        [0.0, 3e-2 - 1j * 5e-3, 1e-3, -3e-2 - 1j * 5e-3, 2e-4 + 1j * 1e-4, 1j * 3e-3, 1e-2, 1j * 3e-3, 2e-4 - 1j * 1e-4]
    )
    abd_target = abd.transform(
        supertranslation=supertranslation,
    )

    abd_recovered, transformation, rel_errs = abd.map_to_superrest_frame(
        t_0=0,
        padding_time=20,
        target_PsiM_input=abd_target.supermomentum("Moreschi"),
        N_itr_maxes={
            "superrest": 1,
            "CoM_transformation": 10,
            "rotation": 10,
            "supertranslation": 10,
        },
    )

    PsiM_diff = abd_recovered.supermomentum("Moreschi")[np.argmin(abs(abd_recovered.t))]
    PsiM_diff -= abd_target.supermomentum("Moreschi")[np.argmin(abs(abd_recovered.t))]
    PsiM_diff[0:4] = 0
    PsiM_diff_S2 = spinsfast.salm2map(PsiM_diff.view(np.ndarray), 0, ell_max, 2 * ell_max + 1, 2 * ell_max + 1)
    PsiM_diff_norm = spinsfast.map2salm((abs(PsiM_diff_S2) ** 2).view(np.ndarray), 0, ell_max)[0] / np.sqrt(4 * np.pi)
    assert np.allclose(PsiM_diff_norm, 0.0, atol=tolerance, rtol=tolerance)

    chi_recovered = abd_recovered.bondi_dimensionless_spin()
    chi_recovered = chi_recovered / np.linalg.norm(chi_recovered, axis=-1)[:, None]
    assert np.allclose(chi_recovered, [[0, 0, 1]] * abd_recovered.t.size, atol=tolerance, rtol=tolerance)

    G = abd_recovered.bondi_CoM_charge() / abd_recovered.bondi_four_momentum()[:, 0, None]
    assert np.allclose(G[np.argmin(abs(abd_recovered.t))], 0.0, atol=tolerance, rtol=tolerance)
    assert np.allclose(
        derivative(G, abd_recovered.t)[np.argmin(abs(abd_recovered.t))], 0.0, atol=tolerance, rtol=tolerance
    )
